local L = AceLibrary("AceLocale-2.2"):new("RaidBuffStatus")
local options = { 
	type='group',
	args = {
		show = {
			type = 'execute',
			name = 'show',
			desc = L["Show the buff report summary window."],
			usage = "show",
			func = function()
				RaidBuffStatus:DoReport()
				RaidBuffStatus:ShowReportFrame()
			end,
		},
		hide = {
			type = 'execute',
			name = 'hide',
			desc = L["Hide the buff report summary window."],
			usage = "hide",
			func = "HideReportFrame",
		},
		toggle = {
			type = 'execute',
			name = 'toggle',
			desc = L["Hide and show the buff report summary window."],
			usage = "toggle",
			func = function()
				if RaidBuffStatus.frame then
					if RaidBuffStatus.frame:IsVisible() then
						RaidBuffStatus:HideReportFrame()
					else
						RaidBuffStatus:DoReport()
						RaidBuffStatus:ShowReportFrame()
					end
				end
			end,
		},
		report = {
			type = 'execute',
			name = 'report',
			desc = L["Report to /raid or /party who is not buffed to the max."],
			usage = "report",
			func = function()
				RaidBuffStatus:DoReport()
				RaidBuffStatus:ReportToChat(true)
			end,
		},
    	},
}


RaidBuffStatus = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0","AceDB-2.0","AceEvent-2.0","AceDebug-2.0")
local Aura = AceLibrary("SpecialEvents-Aura-2.0")
local G = AceLibrary("Gratuity-2.0")

local addon = RaidBuffStatus

addon:RegisterChatCommand("/rbs",options)
addon:RegisterChatCommand("/raidbuffstatus",options)
addon:RegisterDB("RaidBuffStatusDB")

--addon:SetDebugging(true)

-- babblespell replacement using GetSpellInfo(key)
local BSmeta = {
	__index = function(self, key)
		local name, icon
		if type(key) == "number" then
			name, _, icon = GetSpellInfo(key)
		elseif type(key) == "string" then
			local keyfromname = BSK[key]
			if keyfromname then
				name, _, icon = GetSpellInfo(keyfromname)
			else
				geterrorhandler()(("Unknown spell key for name %q"):format(key))
			end
		else
			geterrorhandler()(("Unknown spell key %q"):format(key))
		end
		if name then
			BS[key] = name
			BS[name] = name
			BSI[key] = icon
			BSI[name] = icon
		else
			BS[key] = false
			BSI[key] = false
			geterrorhandler()(("Unknown spell info key %q"):format(key))
		end
		return self[key]
	end,
}
BS = setmetatable({}, BSmeta)
BSI = setmetatable({}, BSmeta)

BSK = {
	["Power Word: Fortitude"] = 1243,
	["Prayer of Fortitude"] = 21562,
	["Mark of the Wild"] = 1126,
	["Gift of the Wild"] = 21849,
	["Arcane Intellect"] = 1459,
	["Arcane Brilliance"] = 23028,
	["Divine Spirit"] = 14752,
	["Prayer of Spirit"] = 27681,
	["Shadow Protection"] = 976,
	["Prayer of Shadow Protection"] = 27683,
	["Crusader Aura"] = 32223,
	["Devotion Aura"] = 465,
	["Retribution Aura"] = 7294,
	["Concentration Aura"] = 19746,
	["Sanctity Aura"] = 20218,
	["Shadow Resistance Aura"] = 19876,
	["Frost Resistance Aura"] = 19888,
	["Fire Resistance Aura"] = 19891,
	["Aspect of the Monkey"] = 13163,
	["Aspect of the Hawk"] = 13165,
	["Aspect of the Beast"] = 13161,
	["Aspect of the Wild"] = 20043,
	["Aspect of the Viper"] = 34074,
	["Aspect of the Cheetah"] = 5118,
	["Aspect of the Pack"] = 13159,
	["Mage Armor"] = 6117,
	["Frost Armor"] = 168,
	["Ice Armor"] = 7302,
	["Molten Armor"] = 30482,
	["Righteous Fury"] = 25780,
	["Soulstone Resurrection"] = 20707,
	["Inner Fire"] = 588,
	["Fel Armor"] = 28176,
	["Omen of Clarity"] = 16864,
	["Lightwell"] = 724,
	["Well Fed"] = 35272,
}

local function SpellName(spellID)
	local name = GetSpellInfo(spellID)
	return name
end

-- Thanks to BigBrother developers saving me time looking these up!

local flasks = {
	SpellName(17626), -- 17626 Flask of the Titans
	SpellName(17627), -- 17627 [Flask of] Distilled Wisdom
	SpellName(17628), -- 17628 [Flask of] Supreme Power
	SpellName(17629), -- 17629 [Flask of] Chromatic Resistance
	SpellName(28518), -- 28518 Flask of Fortification
	SpellName(28519), -- 28519 Flask of Mighty Restoration
	SpellName(28520), -- 28520 Flask of Relentless Assault
	SpellName(28521), -- 28521 Flask of Blinding Light
	SpellName(28540), -- 28540 Flask of Pure Death
	SpellName(33053), -- 33053 Mr. Pinchy's Blessing
	SpellName(42735), -- 42735 [Flask of] Chromatic Wonder
	SpellName(40567), -- 40567 Unstable Flask of the Bandit
	SpellName(40568), -- 40568 Unstable Flask of the Elder
	SpellName(40572), -- 40572 Unstable Flask of the Beast
	SpellName(40573), -- 40573 Unstable Flask of the Physician
	SpellName(40575), -- 40575 Unstable Flask of the Soldier
	SpellName(40576), -- 40576 Unstable Flask of the Sorcerer
	SpellName(41608), -- 41608 Relentless Assault of Shattrath
	SpellName(41609), -- 41609 Fortification of Shattrath
	SpellName(41610), -- 41610 Mighty Restoration of Shattrath
	SpellName(41611), -- 41611 Sureme Power of Shattrath
	SpellName(46837), -- 46837 Pure Death of Shattrath
	SpellName(46839), -- 46839 Blinding Light of Shattrath
}


local belixirs = {
	SpellName(11390),-- 11390 Arcane Elixir
	SpellName(17538),-- 17538 Elixir of the Mongoose
	SpellName(17539),-- 17539 Greater Arcane Elixir
	SpellName(28490),-- 28490 Major Strength
	SpellName(28491),-- 28491 Healing Power
	SpellName(28493),-- 28493 Major Frost Power
	SpellName(28497),-- 28497 Major Agility
	SpellName(28501),-- 28501 Major Firepower
	SpellName(28503),-- 28503 Major Shadow Power
	SpellName(38954),-- 38954 Fel Strength Elixir
	SpellName(33720),-- 33720 Onslaught Elixir
	SpellName(33721),-- 33721 Adept's Elixir
	SpellName(33726),-- 33726 Elixir of Mastery
	SpellName(26276),-- 26276 Elixir of Greater Firepower
	SpellName(45373),-- 45373 Bloodberry - only works on Sunwell Plateau
}
local gelixirs = {
	SpellName(11348),-- 11348 Greater Armor/Elixir of Superior Defense
	SpellName(11396),-- 11396 Greater Intellect
	SpellName(24363),-- 24363 Mana Regeneration/Mageblood Potion
	SpellName(28502),-- 28502 Major Armor/Elixir of Major Defense
	SpellName(28509),-- 28509 Greater Mana Regeneration/Elixir of Major Mageblood
	SpellName(28514),-- 28514 Empowerment
	SpellName(29626),-- 29626 Earthen Elixir
	SpellName(39625),-- 39625 Elixir of Major Fortitude
	SpellName(39627),-- 39627 Elixir of Draenic Wisdom
	SpellName(39628),-- 39628 Elixir of Ironskin
}

local foods = {
	SpellName(35272), -- Well Fed
	SpellName(44106), -- "Well Fed" from Brewfest
	SpellName(43730), -- Electrified
	SpellName(43722), -- Enlightened
	SpellName(25661), -- Increased Stamina
	SpellName(21151), -- Rumsey Rum Black Label
}

local fortitude = {
	BS["Power Word: Fortitude"],
	BS["Prayer of Fortitude"],
}

local wild = {
	BS["Mark of the Wild"],
	BS["Gift of the Wild"],
}

local intellect = {
	BS["Arcane Intellect"],
	BS["Arcane Brilliance"],
}

local spirit = {
	BS["Divine Spirit"],
	BS["Prayer of Spirit"],
}

local shadow = {
	BS["Shadow Protection"],
	BS["Prayer of Shadow Protection"],
}

local auras = {
	BS["Crusader Aura"],
	BS["Devotion Aura"],
	BS["Retribution Aura"],
	BS["Concentration Aura"],
	BS["Sanctity Aura"],
	BS["Shadow Resistance Aura"],
	BS["Frost Resistance Aura"],
	BS["Fire Resistance Aura"],
}

local aspects = {
	BS["Aspect of the Monkey"],
	BS["Aspect of the Hawk"],
	BS["Aspect of the Beast"],
	BS["Aspect of the Wild"],
	BS["Aspect of the Viper"],
	BS["Aspect of the Cheetah"],
	BS["Aspect of the Pack"],
}

local badaspects = {
	BS["Aspect of the Cheetah"],
	BS["Aspect of the Pack"],
}

local magearmors = {
	BS["Mage Armor"],
	BS["Frost Armor"],
	BS["Ice Armor"],
	BS["Molten Armor"],
}


local blessings = {
	BS[19742], -- Blessing of Wisdom
	BS[25894], -- Greater Blessing of Wisdom
	BS[19740], -- Blessing of Might
	BS[25782], -- Greater Blessing of Might
	BS[1038], -- Blessing of Salvation
	BS[25895], -- Greater Blessing of Salvation
	BS[19977], -- Blessing of Light
	BS[25890], -- Greater Blessing of Light
	BS[20217], -- Blessing of Kings
	BS[25898], -- Greater Blessing of Kings
	BS[27168], -- Blessing of Sanctuary
	BS[27169], -- Greater Blessing of Sanctuary
}

local blessingsns = {
	BS[19742], -- Blessing of Wisdom
	BS[25894], -- Greater Blessing of Wisdom
	BS[19740], -- Blessing of Might
	BS[25782], -- Greater Blessing of Might
	BS[19977], -- Blessing of Light
	BS[25890], -- Greater Blessing of Light
	BS[20217], -- Blessing of Kings
	BS[25898], -- Greater Blessing of Kings
	BS[27168], -- Blessing of Sanctuary
	BS[27169], -- Greater Blessing of Sanctuary
}

local wisdom = {
	BS[19742], -- Blessing of Wisdom
	BS[25894], -- Greater Blessing of Wisdom
}

local might = {
	BS[19740], -- Blessing of Might
	BS[25782], -- Greater Blessing of Might
}

local flaskzones = {
	gruul = {
		zones = {
			L["Gruul's Lair"],
		},
		flasks = {
			SpellName(40567), -- 40567 Unstable Flask of the Bandit
			SpellName(40568), -- 40568 Unstable Flask of the Elder
			SpellName(40572), -- 40572 Unstable Flask of the Beast
			SpellName(40573), -- 40573 Unstable Flask of the Physician
			SpellName(40575), -- 40575 Unstable Flask of the Soldier
			SpellName(40576), -- 40576 Unstable Flask of the Sorcerer
		},
	},
	shattrath = {
		zones = {
			L["Tempest Keep"],
			L["Serpentshrine Cavern"],
			L["Black Temple"],
			L["Sunwell Plateau"],
			L["Hyjal Summit"],
		},
		flasks = {
			SpellName(41608), -- 41608 Relentless Assault of Shattrath
			SpellName(41609), -- 41609 Fortification of Shattrath
			SpellName(41610), -- 41610 Mighty Restoration of Shattrath
			SpellName(41611), -- 41611 Sureme Power of Shattrath
			SpellName(46837), -- 46837 Pure Death of Shattrath
			SpellName(46839), -- 46839 Blinding Light of Shattrath
		},
	},
}

-- Thanks to MoBuffs for the original list
local wepbuffs = {
	L["(Ward of Shielding)"], -- (Lesser|Greater) Ward of Shielding
	L["^(Weighted %(%+%d+)"], -- Weighted (+xx 
	L["^(Sharpened %(%+%d+)"], -- Sharpened (+xx 
	L["( Poison ?[IVX]*)"], -- Anesthetic Poison, Deadly Poison [IVX]*, Crippling Poison [IVX]*, Wound Poison [IVX]*, Instant Poison [IVX]*, Mind-numbing Poison [IVX]*
	L["(Mana Oil)"], -- (Minor|Superior|Brillant|Blessed) Mana Oil
	L["(Wizard Oil)"], -- (Minor|Superior|Brillant|Blessed) Wizard Oil
	L["(Frost Oil)"], -- Frost Oil
	L["(Shadow Oil)"], -- Shadow Oil
}

local classes = {
	"PRIEST",
	"DRUID",
	"PALADIN",
	"ROGUE",
	"MAGE",
	"WARLOCK",
	"SHAMAN",
	"WARRIOR",
	"HUNTER",
}


raid = {
	classes = {
		PRIEST = {},
		DRUID = {},
		PALADIN = {},
		ROGUE = {},
		MAGE = {},
		WARLOCK = {},
		SHAMAN = {},
		WARRIOR = {},
		HUNTER = {},
	},
	SpiritTalent = {},
	KingsTalent = {},
	SanctuaryTalent = {},
	ClassNumbers = {},
	BuffTimers = {},
	israid = false,
	isparty = false,
	readid = 0,
	size = 1,
}
raid.reset = function()
	raid.classes = {}
	raid.classes.PRIEST = {}
	raid.classes.DRUID = {}
	raid.classes.PALADIN = {}
	raid.classes.ROGUE = {}
	raid.classes.MAGE = {}
	raid.classes.WARLOCK = {}
	raid.classes.SHAMAN = {}
	raid.classes.WARRIOR = {}
	raid.classes.HUNTER = {}
	raid.SpiritTalent = {}
	raid.KingsTalent = {}
	raid.SanctuaryTalent = {}
	raid.ClassNumbers = {}
	raid.BuffTimers = {}
	raid.israid = false
	raid.isparty = false
	raid.readid = 0
	raid.size = 1
	addon:Debug("raid reset")
end


local specicons = {
	UNKNOWN = "Interface\\Icons\\INV_Misc_QuestionMark",
	Hybrid = "Interface\\Icons\\Spell_Nature_ElementalAbsorption",
}

local classicons = {
	PALADIN = "Interface\\Icons\\Ability_ThunderBolt",
	PRIEST = "Interface\\Icons\\INV_Staff_30",
	DRUID = "Interface\\Icons\\Ability_Druid_Maul",
	MAGE = "Interface\\Icons\\INV_Staff_13",
	ROGUE = "Interface\\Icons\\INV_ThrowingKnife_04",
	WARLOCK = "Interface\\Icons\\Spell_Nature_FaerieFire",
	SHAMAN = "Interface\\Icons\\Spell_Nature_BloodLust",
	WARRIOR = "Interface\\Icons\\INV_Sword_27",
	HUNTER = "Interface\\Icons\\INV_Weapon_Bow_07",
	UNKNOWN = "Interface\\Icons\\INV_Misc_QuestionMark",
}


local tfi = {
	namewidth = 120,
	classwidth = 50,
	specwidth = 50,
	specialisationswidth = 120,
	gap = 2,
	edge = 5,
	inset = 3,
	topedge = 45,
	rowheight = 17,
	rowgap = 0,
	maxrows = 40,
	okbuttonheight = 55,
	rowdata = {},
	rowframes = {},
	buttonsize = 15,
	sort = "class",
	sortorder = 1,
}
tfi.namex = tfi.edge
tfi.classx = tfi.namex + tfi.namewidth + tfi.gap
tfi.specx = tfi.classx + tfi.classwidth + tfi.gap
tfi.specialisationsx = tfi.specx + tfi.specwidth + tfi.gap
tfi.framewidth = tfi.specialisationsx + tfi.specialisationswidth + tfi.edge
tfi.rowwidth = tfi.framewidth - tfi.edge - tfi.edge - tfi.inset - tfi.inset


SP = {
	spirittalent = {
		order = 1000,
		icon = BSI["Divine Spirit"],
		tip = L["Can buff Divine Spirit"],
		callalways = true,
		code = function()
			raid.SpiritTalent = {}
			for name,_ in pairs(raid.classes.PRIEST) do
				if raid.classes.PRIEST[name].talents then
					if raid.classes.PRIEST[name].talents.tree[1].talent[14] > 0 then
						table.insert(raid.SpiritTalent, name)
						raid.classes.PRIEST[name].talents.specialisations.spirittalent = true
					end
				end
			end
		end,
	},
	improvedspiritone = {
		order = 990,
		icon = BSI["Divine Spirit"],
		tip = L["Can buff improved Divine Spirit level 1"],
		callalways = true,
		code = function()
			for name,_ in pairs(raid.classes.PRIEST) do
				if raid.classes.PRIEST[name].talents then
					if raid.classes.PRIEST[name].talents.tree[1].talent[15] > 0 then
						raid.classes.PRIEST[name].talents.specialisations.improvedspiritone = true
					end
				end
			end
		end,
	},
	improvedspirittwo = {
		order = 980,
		icon = BSI["Divine Spirit"],
		tip = L["Can buff improved Divine Spirit level 2"],
		callalways = true,
		code = function()
			for name,_ in pairs(raid.classes.PRIEST) do
				if raid.classes.PRIEST[name].talents then
					if raid.classes.PRIEST[name].talents.tree[1].talent[15] > 1 then
						raid.classes.PRIEST[name].talents.specialisations.improvedspirittwo = true
					end
				end
			end
		end,
	},
	improvedmotwone = {
		order = 970,
		icon = BSI["Mark of the Wild"],
		tip = L["Can buff improved Mark of the Wild level 1"],
		callalways = true,
		code = function()
			for name,_ in pairs(raid.classes.DRUID) do
				if raid.classes.DRUID[name].talents then
					if raid.classes.DRUID[name].talents.tree[3].talent[1] > 0 then
						raid.classes.DRUID[name].talents.specialisations.improvedmotwone = true
					end
				end
			end
		end,
	},
	improvedmotwtwo = {
		order = 960,
		icon = BSI["Mark of the Wild"],
		tip = L["Can buff improved Mark of the Wild level 2"],
		callalways = true,
		code = function()
			for name,_ in pairs(raid.classes.DRUID) do
				if raid.classes.DRUID[name].talents then
					if raid.classes.DRUID[name].talents.tree[3].talent[1] > 1 then
						raid.classes.DRUID[name].talents.specialisations.improvedmotwtwo = true
					end
				end
			end
		end,
	},
	improvedmotwthree = {
		order = 950,
		icon = BSI["Mark of the Wild"],
		tip = L["Can buff improved Mark of the Wild level 3"],
		callalways = true,
		code = function()
			for name,_ in pairs(raid.classes.DRUID) do
				if raid.classes.DRUID[name].talents then
					if raid.classes.DRUID[name].talents.tree[3].talent[1] > 2 then
						raid.classes.DRUID[name].talents.specialisations.improvedmotwthree = true
					end
				end
			end
		end,
	},
	improvedmotwfour = {
		order = 940,
		icon = BSI["Mark of the Wild"],
		tip = L["Can buff improved Mark of the Wild level 4"],
		callalways = true,
		code = function()
			for name,_ in pairs(raid.classes.DRUID) do
				if raid.classes.DRUID[name].talents then
					if raid.classes.DRUID[name].talents.tree[3].talent[1] > 3 then
						raid.classes.DRUID[name].talents.specialisations.improvedmotwfour = true
					end
				end
			end
		end,
	},
	improvedmotwfive = {
		order = 930,
		icon = BSI["Mark of the Wild"],
		tip = L["Can buff improved Mark of the Wild level 5"],
		callalways = true,
		code = function()
			for name,_ in pairs(raid.classes.DRUID) do
				if raid.classes.DRUID[name].talents then
					if raid.classes.DRUID[name].talents.tree[3].talent[1] > 4 then
						raid.classes.DRUID[name].talents.specialisations.improvedmotwfive = true
					end
				end
			end
		end,
	},

	improvedhealthstoneone = {
		order = 920,
		icon = "Interface\\Icons\\INV_Stone_04",
		tip = L["Improved Health Stone level 1"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.WARLOCK) do
				if raid.classes.WARLOCK[name].talents then
					if raid.classes.WARLOCK[name].talents.tree[2].talent[1] > 0 then
						raid.classes.WARLOCK[name].talents.specialisations.improvedhealthstoneone = true
					end
				end
			end
		end,
	},
	improvedhealthstonetwo = {
		order = 920,
		icon = "Interface\\Icons\\INV_Stone_04",
		tip = L["Improved Health Stone level 2"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.WARLOCK) do
				if raid.classes.WARLOCK[name].talents then
					if raid.classes.WARLOCK[name].talents.tree[2].talent[1] > 1 then
						raid.classes.WARLOCK[name].talents.specialisations.improvedhealthstonetwo = true
					end
				end
			end
		end,
	},
	lightwell = {
		order = 910,
		icon = BSI["Lightwell"],
		tip = L["Can create a Lightwell"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.PRIEST) do
				if raid.classes.PRIEST[name].talents then
					if raid.classes.PRIEST[name].talents.tree[2].talent[18] > 0 then
						raid.classes.PRIEST[name].talents.specialisations.lightwell = true
					end
				end
			end
		end,
	},
	improvedfortitudeone = {
		order = 900,
		icon = BSI["Power Word: Fortitude"],
		tip = L["Can buff improved Power Word: Fortitude level 1"],
		callalways = true,
		code = function()
			for name,_ in pairs(raid.classes.PRIEST) do
				if raid.classes.PRIEST[name].talents then
					if raid.classes.PRIEST[name].talents.tree[1].talent[4] > 0 then
						raid.classes.PRIEST[name].talents.specialisations.improvedfortitudeone = true
					end
				end
			end
		end,
	},
	improvedfortitudetwo = {
		order = 890,
		icon = BSI["Power Word: Fortitude"],
		tip = L["Can buff improved Power Word: Fortitude level 2"],
		callalways = true,
		code = function()
			for name,_ in pairs(raid.classes.PRIEST) do
				if raid.classes.PRIEST[name].talents then
					if raid.classes.PRIEST[name].talents.tree[1].talent[4] > 1 then
						raid.classes.PRIEST[name].talents.specialisations.improvedfortitudetwo = true
					end
				end
			end
		end,
	},
	improveddemoone = {
		order = 880,
		icon = BSI[1160], -- Demoralizing Shout
		tip = L["Can buff improved Demoralizing Shout level 1"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.WARRIOR) do
				if raid.classes.WARRIOR[name].talents then
					if raid.classes.WARRIOR[name].talents.tree[2].talent[3] > 0 then
						raid.classes.WARRIOR[name].talents.specialisations.improveddemoone = true
					end
				end
			end
		end,
	},
	improveddemotwo = {
		order = 870,
		icon = BSI[1160], -- Demoralizing Shout
		tip = L["Can buff improved Demoralizing Shout level 2"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.WARRIOR) do
				if raid.classes.WARRIOR[name].talents then
					if raid.classes.WARRIOR[name].talents.tree[2].talent[3] > 1 then
						raid.classes.WARRIOR[name].talents.specialisations.improveddemotwo = true
					end
				end
			end
		end,
	},
	improveddemothree = {
		order = 860,
		icon = BSI[1160], -- Demoralizing Shout
		tip = L["Can buff improved Demoralizing Shout level 3"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.WARRIOR) do
				if raid.classes.WARRIOR[name].talents then
					if raid.classes.WARRIOR[name].talents.tree[2].talent[3] > 2 then
						raid.classes.WARRIOR[name].talents.specialisations.improveddemothree = true
					end
				end
			end
		end,
	},
	improveddemofour = {
		order = 850,
		icon = BSI[1160], -- Demoralizing Shout
		tip = L["Can buff improved Demoralizing Shout level 4"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.WARRIOR) do
				if raid.classes.WARRIOR[name].talents then
					if raid.classes.WARRIOR[name].talents.tree[2].talent[3] > 3 then
						raid.classes.WARRIOR[name].talents.specialisations.improveddemofour = true
					end
				end
			end
		end,
	},
	improveddemofive = {
		order = 840,
		icon = BSI[1160], -- Demoralizing Shout
		tip = L["Can buff improved Demoralizing Shout level 5"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.WARRIOR) do
				if raid.classes.WARRIOR[name].talents then
					if raid.classes.WARRIOR[name].talents.tree[2].talent[3] > 4 then
						raid.classes.WARRIOR[name].talents.specialisations.improveddemofive = true
					end
				end
			end
		end,
	},
	blessingofkings = {
		order = 830,
		icon = BSI[20217], -- Blessing of Kings
		tip = L["Can buff Blessing of Kings"],
		callalways = true,
		code = function()
			raid.KingsTalent = {}
			for name,_ in pairs(raid.classes.PALADIN) do
				if raid.classes.PALADIN[name].talents then
					if raid.classes.PALADIN[name].talents.tree[2].talent[6] > 0 then
						table.insert(raid.KingsTalent, name)
						raid.classes.PALADIN[name].talents.specialisations.blessingofkings = true
					end
				end
			end
		end,
	},
	blessingofsanctuary = {
		order = 820,
		icon = BSI[27168], -- Blessing of Sanctuary
		tip = L["Can buff Blessing of Sanctuary"],
		callalways = true,
		code = function()
			raid.SanctuaryTalent = {}
			for name,_ in pairs(raid.classes.PALADIN) do
				if raid.classes.PALADIN[name].talents then
					if raid.classes.PALADIN[name].talents.tree[2].talent[14] > 0 then
						table.insert(raid.SanctuaryTalent, name)
						raid.classes.PALADIN[name].talents.specialisations.blessingofsanctuary = true
					end
				end
			end
		end,
	},
}


BF = {
	pvp = {											-- button name
		order = 1000,
		list = "pvplist",								-- list name
		check = "checkpvp",								-- check name
		default = true,									-- default state enabled
		defaultbuff = false,								-- default state report as buff missing
		defaultwarning = true,								-- default state report as warning
		defaultdash = true,								-- default state show on dash
		defaultdashcombat = false,							-- default state show on dash when in combat
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = true,								-- check when unit is not in this zone
		selfbuff = true,								-- is it a buff the player themselves can fix
		timer = true,									-- rbs will count how many minutes this buff has been missing/active
		chat = L["PVP On"],								-- chat report
		pre = nil,
		main = function(self, name, class, unit, raid, report)				-- called in main loop
			if UnitIsPVP(unit.unitid) then
				table.insert(report.pvplist, name)
--				addon:Debug("is pvp")
			end
		end,
		post = nil,									-- called after main loop
		icon = "Interface\\Icons\\INV_BannerPVP_02",					-- icon
		update = function(self)								-- icon text
			addon:DefaultButtonUpdate(self, report.pvplist, addon.db.profile.checkpvp, true)
		end,
		click = function(self, button, down)						-- button click
			addon:ButtonClick(self, button, down, "pvp")
		end,
		tip = function(self)								-- tool tip
			addon:Tooltip(self, L["PVP is On"], report.pvplist, raid.BuffTimers.pvptimerlist)
		end,
		partybuff = nil,
	},
	crusader = {
		order = 990,
		list = "crusaderlist",
		check = "checkcrusader",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = true,
		timer = false,
		chat = BS["Crusader Aura"],
		main = function(self, name, class, unit, raid, report)
			if class == "PALADIN" then
				report.checking.crusader = true
				if Aura:UnitHasBuff(unit.unitid, BS["Crusader Aura"]) then
					table.insert(report.crusaderlist, name)
--					addon:Debug("crusader")
				end
			end
		end,
		post = nil,
		icon = BSI["Crusader Aura"],
		update = function(self)
			addon:DefaultButtonUpdate(self, report.crusaderlist, addon.db.profile.checkcrusader, report.checking.crusader or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "crusader")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Paladin has Crusader Aura"], report.crusaderlist)
		end,
		partybuff = nil,
	},

	shadows = {
		order = 980,
		list = "shadowslist",
		check = "checkshadows",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = true,
		timer = false,
		chat = L["Shadow Resistance Aura AND Shadow Protection"],
		main = function(self, name, class, unit, raid, report)
			if raid.ClassNumbers.PRIEST > 0 then
				if class == "PALADIN" then
					report.checking.shadows = true
					if Aura:UnitHasBuff(unit.unitid, BS["Shadow Resistance Aura"]) then
						table.insert(report.shadowslist, name)
--						addon:Debug("2 shadow")
					end
				end
			end
		end,
		post = nil,
		icon = BSI["Shadow Resistance Aura"],
		update = function(self)
			addon:DefaultButtonUpdate(self, report.shadowslist, addon.db.profile.checkshadows, report.checking.shadows or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "shadows")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Paladin has Shadow Resistance Aura AND Shadow Protection"], report.shadowslist)
		end,
		partybuff = nil,
	},

	wrongblessing = {
		order = 975,
		list = "wrongblessinglist",
		check = "checkwrongblessing",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = false,
		timer = false,
		chat = L["Wrong Paladin blessing"],
		main = function(self, name, class, unit, raid, report)
			if raid.ClassNumbers.PALADIN < 1 then
				return
			end
			report.checking.wrongblessing = true
			if class == "ROGUE" then
				local hasbuff = false
				for _, v in ipairs(wisdom) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						hasbuff = true
						break
					end
				end
				if hasbuff then
					table.insert(report.wrongblessinglist, name .. "(" .. BS[19742] .. ")" ) -- Blessing of Wisdom
				end
			elseif class == "MAGE" or class == "WARLOCK" or class == "PRIEST" then
				local hasbuff = false
				for _, v in ipairs(might) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						hasbuff = true
						break
					end
				end
				if hasbuff then
					table.insert(report.wrongblessinglist, name .. "(" .. BS[19740] .. ")" ) -- Blessing of Might
				end
			end
		end,
		post = nil,
		icon = BSI[19740], -- Blessing of Might
		update = function(self)
			addon:DefaultButtonUpdate(self, report.wrongblessinglist, addon.db.profile.checkwrongblessing, report.checking.wrongblessing or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "wrongblessing")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Player has a wrong Paladin blessing"], report.wrongblessinglist)
		end,
		partybuff = nil,
	},

	health = {
		order = 970,
		list = "healthlist",
		check = "checkhealth",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultdashcombat = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = true,
		timer = false,
		chat = L["Health less than 80%"],
		main = function(self, name, class, unit, raid, report)
			if not unit.isdead then
				local maxhealth = UnitHealthMax(unit.unitid)
				local health = UnitHealth(unit.unitid)
				local percent = (health/maxhealth)*100
				if percent < 80 then
					table.insert(report.healthlist, name)
--					addon:Debug("health")
				end
			end
		end,
		post = nil,
		icon = "Interface\\Icons\\INV_Potion_131",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.healthlist, addon.db.profile.checkhealth, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "health")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Player has health less than 80%"], report.healthlist)
		end,
		partybuff = nil,
	},

	mana = {
		order = 960,
		list = "manalist",
		check = "checkmana",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultdashcombat = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = true,
		timer = false,
		chat = L["Mana less than 80%"],
		main = function(self, name, class, unit, raid, report)
			if not unit.isdead then
				if class ~= "WARRIOR" and class ~= "ROGUE" then
					local maxmana = UnitManaMax(unit.unitid)
					local mana = UnitMana(unit.unitid)
					local percent = (mana/maxmana)*100
					if percent < 80 then
						table.insert(report.manalist, name)
--						addon:Debug("mana")
					end
				end
			end
		end,
		post = nil,
		icon = "Interface\\Icons\\INV_Potion_137",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.manalist, addon.db.profile.checkmana, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "mana")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Player has mana less than 80%"], report.manalist)
		end,
		partybuff = nil,
	},
	zone = {
		order = 950,
		list = "zonelist",
		check = "checkzone",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultdashcombat = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = true, -- actually has no effect
		selfbuff = false,
		timer = false,
		chat = L["Different Zone"],
		main = nil, -- done by main code
		post = nil,
		icon = "Interface\\Icons\\INV_Misc_QuestionMark",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.zonelist, addon.db.profile.checkzone, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "zone")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Player is in a different zone"], report.zonelist)
		end,
		partybuff = nil,
	},

	offline = {
		order = 940,
		list = "offlinelist",
		check = "checkoffline",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultdashcombat = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = true, -- actualy has no effect
		selfbuff = false,
		timer = true,
		chat = L["Offline"],
		main = nil, -- done by main code
		post = nil,
		icon = "Interface\\Icons\\INV_Gizmo_FelStabilizer",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.offlinelist, addon.db.profile.checkoffline, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "offline")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Player is Offline"], report.offlinelist, raid.BuffTimers.offlinetimerlist)
		end,
		partybuff = nil,
	},

	afk = {
		order = 930,
		list = "afklist",
		check = "checkafk",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = true,
		selfbuff = true,
		timer = true,
		chat = L["AFK"],
		main = function(self, name, class, unit, raid, report)
			if UnitIsAFK(unit.unitid) then
				table.insert(report.afklist, name)
--				addon:Debug("afk")
			end
		end,
		post = nil,
		icon = "Interface\\Icons\\Trade_Fishing",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.afklist, addon.db.profile.checkafk, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "afk")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Player is AFK"], report.afklist, raid.BuffTimers.afktimerlist)
		end,
		partybuff = nil,
	},

	dead = {
		order = 920,
		list = "deadlist",
		check = "checkdead",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultdashcombat = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = true,
		selfbuff = false,
		timer = true,
		chat = L["Dead"],
		main = function(self, name, class, unit, raid, report)
			if unit.isdead then
				table.insert(report.deadlist, name)
--				addon:Debug("dead")
			end
		end,
		post = nil,
		icon = "Interface\\Icons\\Spell_Holy_SenseUndead",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.deadlist, addon.db.profile.checkdead, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "dead")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Player is Dead"], report.deadlist, raid.BuffTimers.deadtimerlist)
		end,
		partybuff = nil,
	},


	auras = {
		order = 910,
		list = "auraslist",
		check = "checkauras",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = false,
		timer = false,
		chat = function(report, raid, prefix)
			prefix = prefix or ""
			for _, v in ipairs(report.auraslist) do
				local palalist = {}
				for i, k in pairs(raid.classes.PALADIN) do
					if k.group == v then
						table.insert(palalist, i)
					end
				end
				addon:Say(prefix .. "<" .. L["Paladin Different Aura - Group"] .. ">: " .. v .. ": " .. table.concat(palalist, ", "))
			end
		end,
		pre = function(self, raid, report)
			report.palapergrouplist = {{},{},{},{},{},{},{},{}}
			report.palaaurapergrouplist = {{},{},{},{},{},{},{},{}}
		end,
		main = function(self, name, class, unit, raid, report)
			if class == "PALADIN" then
				report.checking.auras = true
				table.insert(report.palapergrouplist[unit.group], name)
				for _, v in ipairs(auras) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						local foundauraingroup = false
						for _, k in ipairs(report.palaaurapergrouplist[unit.group]) do
							if v == k then
								foundauraingroup = true
								break
							end
						end
						if not foundauraingroup then
							table.insert(report.palaaurapergrouplist[unit.group], v)
						end
					end
				end
			end
		end,
		post = function(self, raid, report)
			for i, v in ipairs(report.palapergrouplist) do
			if # v > 1 then
				if # v > # report.palaaurapergrouplist[i] then
					table.insert(report.auraslist, i)
--					addon:Debug("not enough aura:" .. i)
				end
			end
		end
		end,
		icon = BSI["Retribution Aura"],
		update = function(self)
			addon:DefaultButtonUpdate(self, report.auraslist, addon.db.profile.checkauras, report.checking.auras or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "auras")
		end,
		tip = function(self)
			addon:Tooltip(self, L["There are more Paladins than different Auras in group"], report.auraslist)
		end,
		partybuff = nil,
	},

	cheetahpack = {
		order = 900,
		list = "cheetahpacklist",
		check = "checkcheetahpack",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = true,
		timer = false,
		chat = L["Aspect Cheetah/Pack On"],
		main = function(self, name, class, unit, raid, report)
			if class == "HUNTER" then
				report.checking.cheetahpack = true
				local hasbuff = false
				for _, v in ipairs(badaspects) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						hasbuff = true
						break
					end
				end
				if hasbuff then
					if addon.db.profile.ReportGroupNumber then
						table.insert(report.cheetahpacklist, name .. "(" .. unit.group .. ")" )
					else
						table.insert(report.cheetahpacklist, name)
					end
				end
			end
		end,
		post = function(self, raid, report)
			addon:SortNameBySuffix(report.cheetahpacklist)
		end,
		icon = BSI["Aspect of the Cheetah"],
		update = function(self)
			addon:DefaultButtonUpdate(self, report.cheetahpacklist, addon.db.profile.checkcheetahpack, report.checking.cheetahpack or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "cheetahpack")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Aspect of the Cheetah or Pack is on"], report.cheetahpacklist)
		end,
		partybuff = nil,
	},

	righteousfury = {
		order = 890,
		list = "righteousfurylist",
		check = "checkrighteousfury",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = true,
		timer = false,
		chat = BS["Righteous Fury"],
		main = function(self, name, class, unit, raid, report)
			if class == "PALADIN" then
				report.checking.righteousfury = true
				if raid.classes.PALADIN[name].talents then
					if raid.classes.PALADIN[name].talents.spec == "Protection" then
						if not Aura:UnitHasBuff(unit.unitid, BS["Righteous Fury"]) then
							table.insert(report.righteousfurylist, name)
						end
					end
				end
			end
		end,
		post = nil,
		icon = BSI["Righteous Fury"],
		update = function(self)
			addon:DefaultButtonUpdate(self, report.righteousfurylist, addon.db.profile.checkrighteousfury, report.checking.righteousfury or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "righteousfury", BS["Righteous Fury"])
		end,
		tip = function(self)
			addon:Tooltip(self, L["Protection paladin with no Righteous Fury"], report.righteousfurylist)
		end,
		partybuff = nil,
	},

--	thorns = {
--		order = 880,
--		list = "thornslist",
--		check = "checkthorns",
--		default = true,
--		defaultbuff = false,
--		defaultwarning = true,
--		defaultdash = true,
--		defaultdashcombat = false,
--		defaultboss = true,
--		defaulttrash = false,
--		checkzonedout = false,
--		selfbuff = false,
--		timer = false,
--		chat = BS[26992],  -- Thorns
--		main = function(self, name, class, unit, raid, report)
--			if raid.ClassNumbers.DRUID < 1 then
--				return
--			end
--			if not unit.istank then
--				return  -- only tanks need thorns
--			end
--			if class == "PALADIN" or class == "DRUID" or class == "WARRIOR" then  -- only mele tanks need thorns
--				report.checking.thorns = true
--				if not Aura:UnitHasBuff(unit.unitid, BS[26992]) then  -- Thorns
--					table.insert(report.thornslist, name)
--				end
--			end
--		end,
--		post = nil,
--		icon = BSI[26992],  -- Thorns
--		update = function(self)
--			addon:DefaultButtonUpdate(self, report.thornslist, addon.db.profile.checkthorns, report.checking.thorns or false)
--		end,
--		click = function(self, button, down)
--			addon:ButtonClick(self, button, down, "thorns", BS[26992], true)  -- Thorns
--		end,
--		tip = function(self)
--			addon:Tooltip(self, L["Tank missing Thorns"], report.thornslist, nil, BF.thorns:buffers())
--		end,
--		partybuff = function(reportl, prefix)
--			for name,_ in pairs(raid.classes.DRUID) do
--				addon:Say(prefix .. "<" .. BF.thorns.chat .. ">: " .. table.concat(reportl, ", "), name)
--			end
--		end,
--		buffers = function()
--			local thedruids = {}
--			for name,_ in pairs(raid.classes.DRUID) do
--				table.insert(thedruids, name)
--			end
--			return thedruids
--		end,
--	},

--	salvation = {
--		order = 870,
--		list = "salvationlist",
--		check = "checksalvation",
--		default = true,
--		defaultbuff = false,
--		defaultwarning = true,
--		defaultdash = true,
--		defaultdashcombat = false,
--		defaultboss = true,
--		defaulttrash = true,
--		checkzonedout = false,
--		selfbuff = true,
--		timer = false,
--		chat = BS[1038],  -- Blessing of Salvation
--		main = function(self, name, class, unit, raid, report)
--			if raid.ClassNumbers.PALADIN < 1 then
--				return
--			end
--			if not unit.istank then
--				return  -- only tanks need not salvation
--			end
--			if class == "PALADIN" or class == "DRUID" or class == "WARRIOR" or class == "MAGE" or class == "WARLOCK" then  -- only tanks
--				report.checking.salvation = true
--				if Aura:UnitHasBuff(unit.unitid, BS[1038]) or Aura:UnitHasBuff(unit.unitid, BS[25895]) then  -- Blessing of Salvation and Greater
--					table.insert(report.salvationlist, name)
--				end
--			end
--		end,
--		post = nil,
--		icon = BSI[25895],  -- Greater Blessing of Salvation
--		update = function(self)
--			addon:DefaultButtonUpdate(self, report.salvationlist, addon.db.profile.checksalvation, report.checking.salvation or false)
--		end,
--		click = function(self, button, down)
--			addon:ButtonClick(self, button, down, "salvation")
--		end,
--		tip = function(self)
--			addon:Tooltip(self, L["Tank with Salvation"], report.salvationlist)
--		end,
--		partybuff = nil,
--	},


	soulstone = {
		order = 860,
		list = "nosoulstonelist",
		check = "checksoulstone",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultdashcombat = true,
		defaultboss = true,
		defaulttrash = false,
		checkzonedout = true,
		selfbuff = false,
		timer = false,
		chat = function(report, raid, prefix)
			prefix = prefix or ""
			if report.checking.soulstone then
				if # report.soulstonelist < 1 then
					addon:Say(prefix .. "<" .. L["No Soulstone detected"] .. ">")
				end
			end
		end,
		pre = function(self, raid, report)
			report.soulstonelist = {}
		end,
		main = function(self, name, class, unit, raid, report)
			if raid.ClassNumbers.WARLOCK > 0 then
				report.checking.soulstone = true
				if Aura:UnitHasBuff(unit.unitid, BS["Soulstone Resurrection"]) then
					table.insert(report.soulstonelist, name)
				end
			end
		end,
		post = function(self, raid, report)
			if raid.ClassNumbers.WARLOCK > 0 then
				if # report.soulstonelist < 1 then
					table.insert(report.nosoulstonelist, "raid")
				end
			end
		end,
		icon = "Interface\\Icons\\Spell_Shadow_SoulGem",
		update = function(self)
			if addon.db.profile.checksoulstone then
				if report.checking.soulstone then
					self:SetAlpha(1)
					if # report.soulstonelist > 0 then
						self.count:SetText("0")
					else
						self.count:SetText("1")
					end
				else
					self:SetAlpha(0.15)
					self.count:SetText("")
				end
			else
				self:SetAlpha(0.5)
				self.count:SetText("X")
			end
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "soulstone")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Someone has a Soulstone"], report.soulstonelist)
		end,
		partybuff = nil,
	},

	food = {
		order = 500,
		list = "foodlist",
		check = "checkfood",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = true,
		timer = false,
		chat = BS["Well Fed"],
		main = function(self, name, class, unit, raid, report)
			local missingbuff = true
			for _, v in ipairs(foods) do
				if Aura:UnitHasBuff(unit.unitid, v) then
					missingbuff = false
					break
				end
			end
			if missingbuff then
				table.insert(report.foodlist, name)
--				addon:Debug("food")
			end
		end,
		post = nil,
		icon = "Interface\\Icons\\INV_Misc_Food_74",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.foodlist, addon.db.profile.checkfood, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "food")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Not Well Fed"], report.foodlist)
		end,
		partybuff = nil,
	},
	flask = {
		order = 490,
		list = "flasklist",
		check = "checkflaskir",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = false,
		checkzonedout = false,
		selfbuff = true,
		timer = false,
		chat = L["Flask or two Elixirs"],
		pre = function(self, raid, report)
			report.belixirlist = {}
			report.gelixirlist = {}
			report.flaskzonelist = {}
		end,
		main = function(self, name, class, unit, raid, report)
			local missingbuff = true
			for _, v in ipairs(flasks) do
				if Aura:UnitHasBuff(unit.unitid, v) then
					missingbuff = false
					-- has flask now check the zone
					if raid.israid then
						local thiszone = GetRealZoneText()
						local flaskmatched = false
						for _, types in pairs (flaskzones) do
							for _, flask in ipairs(types.flasks) do
								if flask == v then
									flaskmatched = true
									local zonematched = false
									for _, zone in ipairs(types.zones) do
										if thiszone == zone then
											zonematched = true
											break
										end
									end
									if not zonematched then
										table.insert(report.flaskzonelist, name .. "(" .. v .. ")")
									end
								break
								end
							end
							if flaskmatched then break end
						end
					end
					break
				end
			end
			if missingbuff then
				local numbbelixir = 0
				local numbgelixir = 0
				for _, v in ipairs(belixirs) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						numbbelixir = 1
						break
					end
				end
				for _, v in ipairs(gelixirs) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						numbgelixir = 1
						break
					end
				end
				local totalelixir = numbbelixir + numbgelixir
				if totalelixir == 0 then
					table.insert(report.flasklist, name) -- no flask or elixir
--					addon:Debug("flask/elix")
				elseif totalelixir == 1 then
--					addon:Debug("elix")
					if numbbelixir == 0 then
						table.insert(report.belixirlist, name)
--						addon:Debug("belix")
					else
						table.insert(report.gelixirlist, name)
--						addon:Debug("gelix")
					end
				end
			end
		end,
		post = nil,
		icon = "Interface\\Icons\\INV_Potion_119",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.flasklist, addon.db.profile.checkflaskir, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "flask")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Missing a Flask or two Elixirs"], report.flasklist)
		end,
		partybuff = nil,
	},
	belixir = {
		order = 480,
		list = "belixirlist",
		check = "checkflaskir",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = false,
		checkzonedout = false,
		selfbuff = true,
		timer = false,
		chat = L["Battle Elixir"],
		pre = nil,
		main = nil,
		post = nil,
		icon = "Interface\\Icons\\INV_Potion_111",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.belixirlist, addon.db.profile.checkflaskir, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "flask")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Missing a Battle Elixir"], report.belixirlist)
		end,
		partybuff = nil,
	},
	
	gelixir = {
		order = 470,
		list = "gelixirlist",
		check = "checkflaskir",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = false,
		checkzonedout = false,
		selfbuff = true,
		timer = false,
		chat = L["Guardian Elixir"],
		pre = nil,
		main = nil,
		post = nil,
		icon = "Interface\\Icons\\INV_Potion_158",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.gelixirlist, addon.db.profile.checkflaskir, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "flask")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Missing a Guardian Elixir"], report.gelixirlist)
		end,
		partybuff = nil,
	},

	flaskzone = {
		order = 465,
		list = "flaskzonelist",
		check = "checkflaskir",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = true,
		timer = false,
		chat = L["Wrong flask for this zone"],
		pre = nil,
		main = nil,
		post = nil,
		icon = "Interface\\Icons\\INV_Potion_35",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.flaskzonelist, addon.db.profile.checkflaskir, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "flaskzone")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Wrong flask for this zone"], report.flaskzonelist)
		end,
		partybuff = nil,
	},
	wepbuff = {
		order = 464,
		list = "wepbufflist",
		check = "checkwepbuff",
		default = false,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = false,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = false,
		checkzonedout = false,
		selfbuff = true,
		timer = false,
		chat = L["Weapon buff"],
		pre = nil,
		main = function(self, name, class, unit, raid, report)
			if _G.InspectFrame and _G.InspectFrame:IsShown() then
				return -- can't inspect at same time as UI
			end
			if inspectqueuename ~= "" then
				return  -- can't inspect as someone in the queue
			end
			if not CanInspect(unit.unitid) then
				return
			end
			NotifyInspect(unit.unitid)
			local missingbuffmh = true
			local missingbuffoh = true
			G:SetInventoryItem(unit.unitid, 16)
			for _,buff in ipairs(wepbuffs) do
				if G:Find(buff) then
					missingbuffmh = false
					break
				end
			end
			G:SetInventoryItem(unit.unitid, 17)
			for _,buff in ipairs(wepbuffs) do
				if G:Find(buff) then
					missingbuffoh = false
					break
				end
			end
			if addon:IsMele(name, class) and addon:InGroupWithEnhShaman(unit) then
				if not missingbuffmh then   -- mele with enhancement shaman should not have wep bufs
					table.insert(report.wepbufflist, name .. L["(Remove buff)"])
				end
			else
				if missingbuffmh and missingbuffoh then
					table.insert(report.wepbufflist, name)
				end
			end
		end,
		post = nil,
		icon = "Interface\\Icons\\INV_Potion_101",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.wepbufflist, addon.db.profile.checkwepbuff, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "wepbuff")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Missing a temporary weapon buff"], report.wepbufflist)
		end,
		partybuff = nil,
	},

	spirit = {
		order = 460,
		list = "spiritlist",
		check = "checkspirit",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultdashcombat = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = false,
		timer = false,
		chat = BS["Divine Spirit"],
		pre = nil,
		main = function(self, name, class, unit, raid, report)
			if # raid.SpiritTalent > 0 then
				report.checking.spirit = true
				if class ~= "ROGUE" and class ~= "WARRIOR" then
					local missingbuff = true
					for _, v in ipairs(spirit) do
						if Aura:UnitHasBuff(unit.unitid, v) then
							missingbuff = false
							break
						end
					end
					if missingbuff then
						if addon.db.profile.ReportGroupNumber then
							table.insert(report.spiritlist, name .. "(" .. unit.group .. ")" )
						else
							table.insert(report.spiritlist, name)
						end
--						addon:Debug("spirit")
					end
				end
			end
		end,
		post = function(self, raid, report)
			addon:SortNameBySuffix(report.spiritlist)
		end,
		icon = BSI["Divine Spirit"],
		update = function(self)
			addon:DefaultButtonUpdate(self, report.spiritlist, addon.db.profile.checkspirit, report.checking.spirit or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "spirit", BS["Divine Spirit"], BS["Prayer of Spirit"], true)
		end,
		tip = function(self)
			addon:Tooltip(self, L["Missing Divine Spirit"], report.spiritlist, nil, BF.spirit:buffers())
		end,
		partybuff = function(reportl, prefix)
			local thepriests = BF.spirit:buffers()
			for _,name in ipairs(thepriests) do
				addon:Say(prefix .. "<" .. BF.spirit.chat .. ">: " .. table.concat(reportl, ", "), name)
			end
		end,
		buffers = function()
			local thepriests = {}
			local maxpoints = 1
			for name,_ in pairs(raid.classes.PRIEST) do
				if raid.classes.PRIEST[name].talents then
					local points = raid.classes.PRIEST[name].talents.tree[1].talent[14] + raid.classes.PRIEST[name].talents.tree[1].talent[15]
					if points > maxpoints then
						maxpoints = points
						thepriests = {}
						table.insert(thepriests, name)
					elseif points == maxpoints then
						table.insert(thepriests, name)
					end
				end
			end
			return thepriests
		end,
	},

	intellect = {
		order = 450,
		list = "intellectlist",
		check = "checkintellect",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultdashcombat = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = false,
		timer = false,
		chat = BS["Arcane Intellect"],
		pre = nil,
		main = function(self, name, class, unit, raid, report)
			if raid.ClassNumbers.MAGE > 0 then
				report.checking.intellect = true
				if class ~= "ROGUE" and class ~= "WARRIOR" then
					local missingbuff = true
					for _, v in ipairs(intellect) do
						if Aura:UnitHasBuff(unit.unitid, v) then
							missingbuff = false
							break
						end
					end
					if missingbuff then
						if addon.db.profile.ReportGroupNumber then
							table.insert(report.intellectlist, name .. "(" .. unit.group .. ")" )
						else
							table.insert(report.intellectlist, name)
						end
--						addon:Debug("int")
					end
				end
			end
		end,
		post = function(self, raid, report)
			addon:SortNameBySuffix(report.intellectlist)
		end,
		icon = BSI["Arcane Intellect"],
		update = function(self)
			addon:DefaultButtonUpdate(self, report.intellectlist, addon.db.profile.checkintellect, report.checking.intellect or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "intellect", BS["Arcane Intellect"], BS["Arcane Brilliance"], true)
		end,
		tip = function(self)
			addon:Tooltip(self, L["Missing Arcane Intellect"], report.intellectlist, nil, BF.intellect:buffers())
		end,
		partybuff = function(reportl, prefix)
			for name,_ in pairs(raid.classes.MAGE) do
				addon:Say(prefix .. "<" .. BF.intellect.chat .. ">: " .. table.concat(reportl, ", "), name)
			end
		end,
		buffers = function()
			local themages = {}
			for name,_ in pairs(raid.classes.MAGE) do
				table.insert(themages, name)
			end
			return themages
		end,
	},

	wild = {
		order = 440,
		list = "wildlist",
		check = "checkwild",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultdashcombat = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = false,
		timer = false,
		chat = BS["Mark of the Wild"],
		pre = nil,
		main = function(self, name, class, unit, raid, report)
			if raid.ClassNumbers.DRUID > 0 then
				report.checking.wild = true
				local missingbuff = true
				for _, v in ipairs(wild) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						missingbuff = false
						break
					end
				end
				if missingbuff then
					if addon.db.profile.ReportGroupNumber then
						table.insert(report.wildlist, name .. "(" .. unit.group .. ")" )
					else
						table.insert(report.wildlist, name)
					end
--					addon:Debug("wild")
				end
			end
		end,
		post = function(self, raid, report)
			addon:SortNameBySuffix(report.wildlist)
		end,
		icon = BSI["Mark of the Wild"],
		update = function(self)
			addon:DefaultButtonUpdate(self, report.wildlist, addon.db.profile.checkwild, report.checking.wild or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "wild", BS["Mark of the Wild"], BS["Gift of the Wild"], true)
		end,
		tip = function(self)
			addon:Tooltip(self, L["Missing Mark of the Wild"], report.wildlist, nil, BF.wild:buffers())
		end,
		partybuff = function(reportl, prefix)
			local thedruids = BF.wild:buffers()
			for _,name in ipairs(thedruids) do
				addon:Say(prefix .. "<" .. BF.wild.chat .. ">: " .. table.concat(reportl, ", "), name)
			end
		end,
		buffers = function()
			local thedruids = {}
			local maxpoints = 0
			for name,_ in pairs(raid.classes.DRUID) do
				if raid.classes.DRUID[name].talents then
					local points = raid.classes.DRUID[name].talents.tree[3].talent[1]
					if points > maxpoints then
						maxpoints = points
						thedruids = {}
						table.insert(thedruids, name)
					elseif points == maxpoints then
						table.insert(thedruids, name)
					end
				else
					table.insert(thedruids, name)
				end
			end
			return thedruids
		end,
	},

	fortitude = {
		order = 430,
		list = "fortitudelist",
		check = "checkfortitude",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultdashcombat = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = BS["Power Word: Fortitude"],
		pre = nil,
		main = function(self, name, class, unit, raid, report)
			if raid.ClassNumbers.PRIEST > 0 then
				report.checking.fortitude = true
				local missingbuff = true
				for _, v in ipairs(fortitude) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						missingbuff = false
						break
					end
				end
				if missingbuff then
					if addon.db.profile.ReportGroupNumber then
						table.insert(report.fortitudelist, name .. "(" .. unit.group .. ")" )
					else
						table.insert(report.fortitudelist, name)
					end
--					addon:Debug("fortitude")
				end
			end
		end,
		post = function(self, raid, report)
			addon:SortNameBySuffix(report.fortitudelist)
		end,
		icon = BSI["Power Word: Fortitude"],
		update = function(self)
			addon:DefaultButtonUpdate(self, report.fortitudelist, addon.db.profile.checkfortitude, report.checking.fortitude or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "fortitude", BS["Power Word: Fortitude"], BS["Prayer of Fortitude"], true)
		end,
		tip = function(self)
			addon:Tooltip(self, L["Missing Power Word: Fortitude"], report.fortitudelist, nil, BF.fortitude:buffers())
		end,
		partybuff = function(reportl, prefix)
			local thepriests = BF.fortitude:buffers()
			for _,name in ipairs(thepriests) do
				addon:Say(prefix .. "<" .. BF.fortitude.chat .. ">: " .. table.concat(reportl, ", "), name)
			end
		end,
		buffers = function()
			local thepriests = {}
			local maxpoints = 0
			for name,_ in pairs(raid.classes.PRIEST) do
				if raid.classes.PRIEST[name].talents then
					local points = raid.classes.PRIEST[name].talents.tree[1].talent[4]
					if points > maxpoints then
						maxpoints = points
						thepriests = {}
						table.insert(thepriests, name)
					elseif points == maxpoints then
						table.insert(thepriests, name)
					end
				else
					table.insert(thepriests, name)
				end
			end
			return thepriests
		end,
	},

	shadow = {
		order = 420,
		list = "shadowlist",
		check = "checkshadow",
		default = false,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultdashcombat = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = false,
		timer = false,
		chat = BS["Shadow Protection"],
		pre = nil,
		main = function(self, name, class, unit, raid, report)
			if raid.ClassNumbers.PRIEST > 0 then
				report.checking.shadow = true
				local missingbuff = true
				for _, v in ipairs(shadow) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						missingbuff = false
						break
					end
				end
				if missingbuff then
					if addon.db.profile.ReportGroupNumber then
						table.insert(report.shadowlist, name .. "(" .. unit.group .. ")" )
					else
						table.insert(report.shadowlist, name)
					end
--					addon:Debug("shadow")
				end
			end
		end,
		post = function(self, raid, report)
			addon:SortNameBySuffix(report.shadowlist)
		end,
		icon = BSI["Shadow Protection"],
		update = function(self)
			addon:DefaultButtonUpdate(self, report.shadowlist, addon.db.profile.checkshadow, report.checking.shadow or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "shadow", BS["Shadow Protection"], BS["Prayer of Shadow Protection"], true)
		end,
		tip = function(self)
			addon:Tooltip(self, L["Missing Shadow Protection"], report.shadowlist, nil, BF.shadow:buffers())
		end,
		partybuff = function(reportl, prefix)
			for name,_ in pairs(raid.classes.PRIEST) do
				addon:Say(prefix .. "<" .. BF.shadow.chat .. ">: " .. table.concat(reportl, ", "), name)
			end
		end,
		buffers = function()
			local thepriests = {}
			for name,_ in pairs(raid.classes.PRIEST) do
				table.insert(thepriests, name)
			end
			return thepriests
		end,
	},

	noaura = {
		order = 410,
		list = "noauralist",
		check = "checknoaura",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = true,
		timer = false,
		chat = L["Paladin Aura"],
		pre = nil,
		main = function(self, name, class, unit, raid, report)
			if class == "PALADIN" then
				report.checking.noaura = true
				local missingbuff = true
				for _, v in ipairs(auras) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						missingbuff = false
						break
					end
				end
				if missingbuff then
					table.insert(report.noauralist, name)
--					addon:Debug("no aura")
				end
			end
		end,
		post = nil,
		icon = BSI["Devotion Aura"],
		update = function(self)
			addon:DefaultButtonUpdate(self, report.noauralist, addon.db.profile.checknoaura, report.checking.noaura or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "noaura")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Paladin has no Aura at all"], report.noauralist)
		end,
		partybuff = nil,
	},

	noaspect = {
		order = 400,
		list = "noaspectlist",
		check = "checknoaspect",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = true,
		timer = false,
		chat = L["Hunter Aspect"],
		pre = nil,
		main = function(self, name, class, unit, raid, report)
			if class == "HUNTER" then
				report.checking.noaspect = true
				local missingbuff = true
				for _, v in ipairs(aspects) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						missingbuff = false
						break
					end
				end
				if missingbuff then
					table.insert(report.noaspectlist, name)
--					addon:Debug("no aspect")

				end
			end
		end,
		post = nil,
		icon = BSI["Aspect of the Monkey"],
		update = function(self)
			addon:DefaultButtonUpdate(self, report.noaspectlist, addon.db.profile.checknoaspect, report.checking.noaspect or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "noaspect")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Hunter has no aspect at all"], report.noaspectlist)
		end,
		partybuff = nil,
	},

	innerfire = {
		order = 390,
		list = "innerfirelist",
		check = "checkinnerfire",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = true,
		timer = false,
		chat = BS["Inner Fire"],
		main = function(self, name, class, unit, raid, report)
			if class == "PRIEST" then
				report.checking.innerfire = true
				if not Aura:UnitHasBuff(unit.unitid, BS["Inner Fire"]) then
					table.insert(report.innerfirelist, name)
--					addon:Debug("innerfire")
				end
			end
		end,
		post = nil,
		icon = BSI["Inner Fire"],
		update = function(self)
			addon:DefaultButtonUpdate(self, report.innerfirelist, addon.db.profile.checkinnerfire, report.checking.innerfire or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "innerfire", BS["Inner Fire"])
		end,
		tip = function(self)
			addon:Tooltip(self, L["Priest is missing Inner Fire"], report.innerfirelist)
		end,
		partybuff = nil,
	},

	felarmor = {
		order = 380,
		list = "felarmorlist",
		check = "checkfelarmor",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = true,
		timer = false,
		chat = BS["Fel Armor"],
		main = function(self, name, class, unit, raid, report)
			if class == "WARLOCK" then
				report.checking.felarmor = true
				if not Aura:UnitHasBuff(unit.unitid, BS["Fel Armor"]) then
					table.insert(report.felarmorlist, name)
--					addon:Debug("felarmor")
				end
			end
		end,
		post = nil,
		icon = BSI["Fel Armor"],
		update = function(self)
			addon:DefaultButtonUpdate(self, report.felarmorlist, addon.db.profile.checkfelarmor, report.checking.felarmor or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "felarmor", BS["Fel Armor"])
		end,
		tip = function(self)
			addon:Tooltip(self, L["Warlock is missing Fel Armor"], report.felarmorlist)
		end,
		partybuff = nil,
	},

	magearmor = {
		order = 370,
		list = "magearmorlist",
		check = "checkmagearmor",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = true,
		timer = false,
		chat = BS["Mage Armor"],
		pre = nil,
		main = function(self, name, class, unit, raid, report)
			if class == "MAGE" then
				report.checking.magearmor = true
				local missingbuff = true
				for _, v in ipairs(magearmors) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						missingbuff = false
						break
					end
				end
				if missingbuff then
					table.insert(report.magearmorlist, name)
--					addon:Debug("magearmor")
				end
			end
		end,
		post = nil,
		icon = BSI["Mage Armor"],
		update = function(self)
			addon:DefaultButtonUpdate(self, report.magearmorlist, addon.db.profile.checkmagearmor, report.checking.magearmor or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "magearmor")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Mage is missing a Mage Armor"], report.magearmorlist)
		end,
		partybuff = nil,
	},

	omenofclarity = {
		order = 360,
		list = "omenofclaritylist",
		check = "checkomenofclarity",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = true,
		timer = false,
		chat = BS["Omen of Clarity"],
		main = function(self, name, class, unit, raid, report)
			if class == "DRUID" then
				if raid.classes.DRUID[name].talents then
					if raid.classes.DRUID[name].talents.tree[3].talent[8] > 0 then
						report.checking.omenofclarity = true
						if not Aura:UnitHasBuff(unit.unitid, BS["Omen of Clarity"]) then
							table.insert(report.omenofclaritylist, name)
--							addon:Debug("omenofclarity")
						end
					end
				end
			end
		end,
		post = nil,
		icon = BSI["Omen of Clarity"],
		update = function(self)
			addon:DefaultButtonUpdate(self, report.omenofclaritylist, addon.db.profile.checkomenofclarity, report.checking.omenofclarity or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "omenofclarity", BS["Omen of Clarity"])
		end,
		tip = function(self)
			addon:Tooltip(self, L["Druid is missing Omen of Clarity"], report.omenofclaritylist)
		end,
		partybuff = nil,
	},

	missingblessing = {
		order = 350,
		list = "missingblessinglist",
		check = "checkmissingblessing",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		selfbuff = false,
		timer = false,
		chat = L["Paladin blessing"],
		pre = function(self, raid, report)
			if raid.ClassNumbers.PALADIN < 1 then
				return
			end
			local commonblessings = 2 -- light, salvation
			if # raid.KingsTalent > 0 then
				commonblessings = commonblessings + 1 -- kings
			end
			if # raid.SanctuaryTalent > 0 then
				commonblessings = commonblessings + 1 -- sanctuary
			end
			local maxblessings = raid.ClassNumbers.PALADIN
			if maxblessings > (commonblessings + 2) then -- might, wisdom
				maxblessings = commonblessings + 2
			end
			report.minblessings = {}
			if maxblessings == 1 then  -- one paladin so everyone will have at least one blessing
				report.minblessings.WARRIOR = 1
				report.minblessings.ROGUE = 1
				report.minblessings.PRIEST = 1
				report.minblessings.MAGE = 1
				report.minblessings.WARLOCK = 1
				report.minblessings.PALADIN = 1
				report.minblessings.DRUID = 1
				report.minblessings["Feral Combat DRUID"] = 1
				report.minblessings.SHAMAN = 1
				report.minblessings.HUNTER = 1
			elseif maxblessings > 1 and maxblessings <= (commonblessings - 1) then
				report.minblessings.WARRIOR = maxblessings - 1  -- might not need salvation
				report.minblessings.ROGUE = maxblessings
				report.minblessings.PRIEST = maxblessings
				report.minblessings.MAGE = maxblessings - 1  -- might not need salvation
				report.minblessings.WARLOCK = maxblessings - 1  -- might not need salvation
				report.minblessings.PALADIN = maxblessings - 1  -- might not need salvation
				report.minblessings.DRUID = maxblessings
				report.minblessings["Feral Combat DRUID"] = maxblessings - 1 -- might not need salvation
				report.minblessings.SHAMAN = maxblessings
				report.minblessings.HUNTER = maxblessings - 1  -- might not need salvation
			else
				report.minblessings.WARRIOR = maxblessings - 1  -- might not need salvation
				report.minblessings.ROGUE = maxblessings - 1  -- does not need wisdom
				report.minblessings.PRIEST = maxblessings - 1 -- does not need might
				report.minblessings.MAGE = maxblessings - 2  -- might not need salvation and does not need might
				report.minblessings.WARLOCK = maxblessings - 2  -- might not need salvation and does not need might
				report.minblessings.PALADIN = maxblessings - 1  -- might not need salvation
				report.minblessings.DRUID = maxblessings - 1 -- does not need might
				report.minblessings["Feral Combat DRUID"] = maxblessings - 1  -- might not need salvation
				report.minblessings.SHAMAN = maxblessings - 1 -- does not need might
				report.minblessings.HUNTER = maxblessings - 1  -- might not need salvation
			end

		end,
		main = function(self, name, class, unit, raid, report)
			if raid.ClassNumbers.PALADIN < 1 then
				return
			end
			report.checking.missingblessing = true
			if raid.ClassNumbers.PALADIN == 1 then  -- special case for 1 pala - everyone gets a blessing
				if addon:NumberOfBlessings(unit.unitid, false) < report.minblessings[class] then
					table.insert(report.missingblessinglist, name)
				end
				return
			end

			if class == "ROGUE" or class == "PRIEST" or class == "SHAMAN" then
				if addon:NumberOfBlessings(unit.unitid, false) < report.minblessings[class] then
					table.insert(report.missingblessinglist, name)
				end
			elseif class == "WARRIOR" or class == "PALADIN" or class == "MAGE" or class == "WARLOCK" or class == "HUNTER" then
				if addon:NumberOfBlessings(unit.unitid, true) < report.minblessings[class] then
					table.insert(report.missingblessinglist, name)
				end
			elseif class == "DRUID" then
				local isferal = false
				if raid.classes.DRUID[name].talents then
					if raid.classes.DRUID[name].talents.spec == "Feral Combat" then
						isferal = true
					end
				end
				if isferal then
					if addon:NumberOfBlessings(unit.unitid, true) < report.minblessings["Feral Combat DRUID"] then
						table.insert(report.missingblessinglist, name)
					end
				else
					if addon:NumberOfBlessings(unit.unitid, false) < report.minblessings.DRUID then
						table.insert(report.missingblessinglist, name)
					end
				end
			end
		end,
		post = nil,
		icon = BSI[25898], -- Greater Blessing of Kings
		update = function(self)
			addon:DefaultButtonUpdate(self, report.missingblessinglist, addon.db.profile.checkmissingblessing, report.checking.missingblessing or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "missingblessing")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Player is missing at least one Paladin blessing"], report.missingblessinglist, nil, BF.missingblessing:buffers())
		end,
		partybuff = function(reportl, prefix)
			for name,_ in pairs(raid.classes.PALADIN) do
				addon:Say(prefix .. "<" .. BF.missingblessing.chat .. ">: " .. table.concat(reportl, ", "), name)
			end
		end,
		buffers = function()
			local thelols = {}
			for name,_ in pairs(raid.classes.PALADIN) do
				table.insert(thelols, name)
			end
			return thelols
		end,
	},

	amplifymagic = {
		order = 340,
		list = "amplifymagiclist",
		check = "checkamplifymagic",
		default = false,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = false,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = false,
		checkzonedout = false,
		selfbuff = false,
		timer = false,
		chat = BS[33946], -- Amplify Magic
		pre = nil,
		main = function(self, name, class, unit, raid, report)
			if raid.ClassNumbers.MAGE > 0 then
				report.checking.amplifymagic = true
				if not Aura:UnitHasBuff(unit.unitid, BS[33946]) then
					table.insert(report.amplifymagiclist, name)
				end
			end
		end,
		post = nil,
		icon = BSI[33946], -- Amplify Magic
		update = function(self)
			addon:DefaultButtonUpdate(self, report.amplifymagiclist, addon.db.profile.checkamplifymagic, report.checking.amplifymagic or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "amplifymagic")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Missing Amplify Magic"], report.amplifymagiclist, nil, BF.amplifymagic:buffers()) -- same as intellect for amplify magic
		end,
		partybuff = function(reportl, prefix)
			for name,_ in pairs(raid.classes.MAGE) do
				addon:Say(prefix .. "<" .. BF.amplifymagic.chat .. ">: " .. table.concat(reportl, ", "), name)
			end
		end,
		buffers = function()
			local themages = {}
			for name,_ in pairs(raid.classes.MAGE) do
				table.insert(themages, name)
			end
			return themages
		end,
	},
	help20080606 = {
		order = 10,
		list = "none",
		check = "checkhelp20080606",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultdashcombat = false,
		defaultboss = true,
		defaulttrash = false,
		checkzonedout = false,
		selfbuff = false,
		timer = false,
		chat = nil,
		pre = nil,
		main = nil,
		post = nil,
		icon = "Interface\\Icons\\Mail_GMIcon",
		update = function(self)
			self.count:SetText("")
		end,
		click = nil,
		tip = function(self)
			GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
			GameTooltip:SetText(L["RBS Dashboard Help"],1,1,1,1,1)
			GameTooltip:AddLine(L["Click buffs to disable and enable."],nil,nil,nil,nil,1)
			GameTooltip:AddLine(L["Shift-Click buffs to report on only that buff."],nil,nil,nil,nil,1)
			GameTooltip:AddLine(L["Ctrl-Click buffs to whisper those who need to buff."],nil,nil,nil,nil,1)
			GameTooltip:AddLine(L["Alt-Click on a self buff will renew that buff."],nil,nil,nil,nil,1)
			GameTooltip:AddLine(L["Alt-Click on a party buff will cast on someone missing that buff."],nil,nil,nil,nil,1)
			GameTooltip:AddLine(L["Remove this button from this dashboard in the options window."],nil,nil,nil,nil,1)
			GameTooltip:Show()
		end,
		partybuff = nil,
	},
}


report = {
	checking = {},
}
report.reset = function()
	for reportname,_ in pairs(report) do
		if type(report[reportname]) == "number" then
			report[reportname] = 0
		elseif type(report[reportname]) == "table" then
			report[reportname] = {}
		end
	end
--	addon:Debug("report reset")
end


local profilelist = {
	ReportSelf = false,
	ReportChat = true,
	ReportGroupNumber = true,
	LockWindow = false,
	IgnoreLastThreeGroups = true,
	x = 0,
	y = 0,
}

for buffcheck, _ in pairs(BF) do
	if BF[buffcheck].list then
		report[BF[buffcheck].list] = {} -- add empty list to report
	end
	if BF[buffcheck].default then  -- if default setting for buff check is enabled
		profilelist[BF[buffcheck].check] = true
	else
		profilelist[BF[buffcheck].check] = false
	end
	for _, defname in ipairs({"buff", "warning", "dash", "dashcombat", "boss", "trash"}) do
		if BF[buffcheck]["default" .. defname] then
			profilelist[buffcheck .. defname] = true
		else
			profilelist[buffcheck .. defname] = false
		end
	end
end


inspectqueuename = ""
local inspectqueueunitid = ""
local inspectqueueclass = ""
local inspectqueuetime = 0
local buttons = {}
local optionsbuttons = {}
local incombat = false
local tankList = '|'

addon:RegisterDefaults("profile", profilelist)


-- End of inits


function addon:OnInitialize()
	if oRA then
		self:RegisterEvent("oRA_MainTankUpdate", "UpdateExternalTankList")
	elseif CT_RA_MainTanks then
		hooksecurefunc("CT_RAOptions_UpdateMTs", function() self:UpdateExternalTankList() end)
	end
end

-- credit for this code goes to Peragor and GridLayoutPlus
function addon:UpdateExternalTankList()
	addon:Debug("Tank list updated")
	-- oRa2 and CT raid integration: get list of unitids for configured tanks
	local tankTable = nil
	tankList = '|'
	if oRA and oRA.maintanktable then
		tankTable = oRA.maintanktable
	elseif CT_RA_MainTanks then
		tankTable = CT_RA_MainTanks
	end
	if tankTable then
		for key, val in pairs(tankTable) do
			addon:Debug("updated tank:" .. val)
			local unitid = addon:GetUnitFromName(val).unitid
			addon:Debug(val)
			if unitid and UnitExists(unitid) and UnitPlayerOrPetInRaid(unitid) then
				tankList = tankList .. val .. '|'
			end
		end
	end
end

function addon:ShowReportFrame()
	ShowUIPanel(RBSFrame)
end

function addon:HideReportFrame()
	HideUIPanel(RBSFrame)
end

function addon:ToggleOptionsFrame()
	if addon.optionsframe:IsVisible() then
		HideUIPanel(RBSOptionsFrame)
	else
		addon:ShowOptionsFrame()
	end
end

function addon:ShowOptionsFrame()
	addon:UpdateOptionsButtons()
	ShowUIPanel(RBSOptionsFrame)
end

function addon:ToggleTalentsFrame()
	if addon.talentframe:IsVisible() then
		HideUIPanel(RBSTalentsFrame)
	else
		addon:ShowTalentsFrame()
	end
end

function addon:ShowTalentsFrame()
	addon:UpdateTalentsFrame()
	ShowUIPanel(RBSTalentsFrame)
end

function addon:UpdateTalentsFrame()
	local height = tfi.topedge + (raid.size * (tfi.rowheight + tfi.rowgap)) + tfi.okbuttonheight
	addon.talentframe:SetHeight(height)
	for i = 1, tfi.maxrows do
		tfi.rowframes[i].rowframe:Hide()
	end
	for speccheck, _ in pairs(SP) do
		SP[speccheck].code()
	end
	addon:GetTalentRowData()
	addon:SortTalentRowData(tfi.sort, tfi.sortorder)
	addon:CopyTalentRowDataToRowFrames()
	for i = 1, raid.size do
		tfi.rowframes[i].rowframe:Show()
	end
end

function addon:GetTalentRowData()
	tfi.rowdata = {}
	local row = 1
	for class,_ in pairs(raid.classes) do
		for name,_ in pairs(raid.classes[class]) do
			local unit = raid.classes[class][name]
			tfi.rowdata[row] = {}
			tfi.rowdata[row].name = name
			tfi.rowdata[row].class = class
			tfi.rowdata[row].specialisations = {}
			if unit.talents then
				tfi.rowdata[row].spec = unit.talents.spec
				tfi.rowdata[row].specicon = unit.talents.specicon
--				addon:Debug(unit.talents.specicon)
				for speccheck, _ in pairs(SP) do
					if unit.talents.specialisations[speccheck] then
						table.insert(tfi.rowdata[row].specialisations, speccheck)
					end
				end
			else
				tfi.rowdata[row].spec = "UNKNOWN"
				tfi.rowdata[row].specicon = "UNKNOWN"
			end
			
			table.sort(tfi.rowdata[row].specialisations, function (a,b)
				return(SP[a].order > SP[b].order)
			end)
			row = row + 1
		end
	end
end

function addon:SortTalentRowData(sort, sortorder)
	tfi.sort = sort
	tfi.sortorder = sortorder
	if sort == "name" then
		table.sort(tfi.rowdata, function (a,b)
			return (addon:Compare(a.name, b.name, sortorder))
		end)
	elseif sort == "class" then
		table.sort(tfi.rowdata, function (a,b)
			if a.class == b.class then
				if a.spec == b.spec then
					return (addon:Compare(a.name, b.name, sortorder))
				end
				return (addon:Compare(a.spec, b.spec, sortorder))
			else
				return (addon:Compare(a.class, b.class, sortorder))
			end
		end)
	elseif sort == "spec" then
		table.sort(tfi.rowdata, function (a,b)
			if a.spec == b.spec then
				return (addon:Compare(a.class, b.class, sortorder))
			else
				return (addon:Compare(a.spec, b.spec, sortorder))
			end
		end)
	elseif sort == "specialisations" then
		table.sort(tfi.rowdata, function (a,b)
			if #a.specialisations == #b.specialisations then
				return (addon:Compare(a.name, b.name, sortorder))
			end
			return (addon:Compare(#a.specialisations, #b.specialisations, sortorder))
		end)
	end
end

function addon:Compare(a, b, sortorder)
	if sortorder == 1 then
			return (a < b)
		else
			return (a > b)
	end
end

function addon:CopyTalentRowDataToRowFrames()
	for i, _ in ipairs(tfi.rowdata) do
		local class = tfi.rowdata[i].class
		local name = tfi.rowdata[i].name
		local r = RAID_CLASS_COLORS[class].r
		local g = RAID_CLASS_COLORS[class].g
		local b = RAID_CLASS_COLORS[class].b
		tfi.rowframes[i].name:SetText(name)
		tfi.rowframes[i].name:SetTextColor(r,g,b)
		tfi.rowframes[i].class:SetNormalTexture(classicons[class] or classicons.UNKNOWN)
		tfi.rowframes[i].class:SetScript("OnEnter", function() addon:Tooltip(tfi.rowframes[i].class, addon:TitleCaps(class), nil) end )
		tfi.rowframes[i].class:SetScript("OnLeave", function() GameTooltip:Hide() end)
		if raid.classes[class][name].talents then
			local spec = raid.classes[class][name].talents.spec
			local specicon = raid.classes[class][name].talents.specicon
			tfi.rowframes[i].spec:SetScript("OnEnter", function() addon:Tooltip(tfi.rowframes[i].spec, spec) end )
			tfi.rowframes[i].spec:SetScript("OnLeave", function() GameTooltip:Hide() end)
			if spec == "Hybrid" then
				tfi.rowframes[i].spec:SetNormalTexture(specicons.Hybrid)
			else
				tfi.rowframes[i].spec:SetNormalTexture(specicon or specicons.UNKNOWN)
			end
		else
			tfi.rowframes[i].spec:SetNormalTexture(specicons.UNKNOWN)
			tfi.rowframes[i].spec:SetScript("OnEnter", function() addon:Tooltip(tfi.rowframes[i].spec, "Unknown") end )
			tfi.rowframes[i].spec:SetScript("OnLeave", function() GameTooltip:Hide() end)
		end
		for j, v in ipairs (tfi.rowframes[i].specialisations) do
			v:Hide()
			local speccheck = tfi.rowdata[i].specialisations[j]
			if speccheck then
				v:SetNormalTexture(SP[speccheck].icon)
				v:SetScript("OnEnter", function() addon:Tooltip(v, SP[speccheck].tip) end )
				v:SetScript("OnLeave", function() GameTooltip:Hide() end)
				v:Show()
			end
		end
	end
end

function addon:DoReport()
	report:reset()
	addon:ReadRaid()
	if (not raid.israid) and (not raid.isparty) then
		addon:Debug("Report aborted as not in raid or party.")
		return
	end
	addon:CalculateReport()
	addon:UpdateButtons()
end

function addon:CalculateReport()
	-- PRE HERE
	for buffcheck, _ in pairs(BF) do
		if addon.db.profile[BF[buffcheck].check] then
			if BF[buffcheck].pre then
--				addon:Debug("pre:" .. buffcheck)
				BF[buffcheck].pre(self, raid, report)
			end
		end
	end

	-- MAIN HERE
	local thiszone = GetRealZoneText()
	for class,_ in pairs(raid.classes) do
		for name,_ in pairs(raid.classes[class]) do
--			addon:Debug("Name:" .. name .. " " .. class)
			local unit = raid.classes[class][name]

			if not unit.online then
				if addon.db.profile.checkoffline then
					table.insert(report.offlinelist, name)  -- used by offline warning check
				end
				break
			end

			local zonedin = true
			if raid.israid then
				if thiszone ~= unit.zone then
					zonedin = false
					if addon.db.profile.checkzone then
						table.insert(report.zonelist, name)
--						addon:Debug("zone")
					end
				end
			end

			for buffcheck, _ in pairs(BF) do
				if addon.db.profile[BF[buffcheck].check] then
					if zonedin or BF[buffcheck].checkzonedout then
						if BF[buffcheck].main then
							BF[buffcheck].main(self, name, class, unit, raid, report)
						end
					end
				end
			end
			if unit.talents == nil then
				addon:RequestInspect(name, class, unit.unitid) -- find out all talent info eventually
			end

		end

	end

	-- do timers
	local thetimenow = math.floor(GetTime())
	for buffcheck, _ in pairs(BF) do
		if BF[buffcheck].timer then
			if not raid.BuffTimers[buffcheck .. "timerlist"] then
				raid.BuffTimers[buffcheck .. "timerlist"] = {}
			end
			for _, v in ipairs(report[BF[buffcheck].list]) do  -- first add those on the list to the timer list if not there
				local missing = true
				for n, t in pairs(raid.BuffTimers[buffcheck .. "timerlist"]) do
					if v == n then
						missing = false
						break
					end
				end
				if missing then
					raid.BuffTimers[buffcheck .. "timerlist"][v] = thetimenow
				end
			end
			for n, t in pairs(raid.BuffTimers[buffcheck .. "timerlist"]) do -- now remove those who are no longer on the list
				local missing = true
				for _, v in ipairs(report[BF[buffcheck].list]) do
					if v == n then
						missing = false
						break
					end
				end
				if missing then
					raid.BuffTimers[buffcheck .. "timerlist"][n] = nil
				end
			end
		end
	end

	-- sort names
	for buffcheck, _ in pairs(BF) do
		if # report[BF[buffcheck].list] > 1 then
			table.sort(report[BF[buffcheck].list])
		end
	end

	-- POST HERE
	for buffcheck, _ in pairs(BF) do
		if addon.db.profile[BF[buffcheck].check] then
			if BF[buffcheck].post then
--				addon:Debug("post:" .. buffcheck)
				BF[buffcheck].post(self, raid, report)
			end
		end
	end
end

function addon:ReportToChat(boss, player)
	local warnings = 0
	local buffs = 0
	for buffcheck, _ in pairs(BF) do
		if # report[BF[buffcheck].list] > 0 then
			if (boss and addon.db.profile[buffcheck .. "boss"]) or ((not boss) and addon.db.profile[buffcheck .. "trash"]) then
				if addon.db.profile[buffcheck .. "warning"] then
					warnings = warnings + # report[BF[buffcheck].list]
				end
				if addon.db.profile[buffcheck .. "buff"] then
					buffs = buffs + # report[BF[buffcheck].list]
				end
			end
		end
	end
	if warnings > 0 then
		addon:Say(L["Warnings: "] .. warnings)
		for buffcheck, _ in pairs(BF) do
			if # report[BF[buffcheck].list] > 0 then
				if (boss and addon.db.profile[buffcheck .. "boss"] ) or ((not boss) and addon.db.profile[buffcheck .. "trash"]) then
					if addon.db.profile[buffcheck .. "warning"] then
						if type(BF[buffcheck].chat) == "string" then
							if BF[buffcheck].timer then
								local timerlist = {}
								for _, n in ipairs(report[BF[buffcheck].list]) do
									if raid.BuffTimers[buffcheck .. "timerlist"][n] then
										table.insert(timerlist, n .. "(" .. addon:TimeSince(raid.BuffTimers[buffcheck .. "timerlist"][n]) .. ")")
									else
										table.insert(timerlist, n)
									end
								end
								addon:Say("<" .. BF[buffcheck].chat .. ">: " .. table.concat(timerlist, ", "))
							else
								addon:Say("<" .. BF[buffcheck].chat .. ">: " .. table.concat(report[BF[buffcheck].list], ", "))
							end
						elseif type(BF[buffcheck].chat) == "function" then
							BF[buffcheck].chat(report, raid)
						end
					end
				end
			end
		end
	end
	if buffs > 0 then
		if boss then
			addon:Say(L["Missing buffs (Boss): "] .. buffs)
		else
			addon:Say(L["Missing buffs (Trash): "] .. buffs)
		end
		for buffcheck, _ in pairs(BF) do
			if # report[BF[buffcheck].list] > 0 then
				if (boss and addon.db.profile[buffcheck .. "boss"] ) or ((not boss) and addon.db.profile[buffcheck .. "trash"]) then
					if addon.db.profile[buffcheck .. "buff"] then
						if type(BF[buffcheck].chat) == "string" then
							addon:Say("<" .. BF[buffcheck].chat .. ">: " .. table.concat(report[BF[buffcheck].list], ", "))
						elseif type(BF[buffcheck].chat) == "function" then
							BF[buffcheck].chat(report, raid)
						end
					end
				end
			end
		end
	else
		if boss then
			addon:Say(L["No buffs needed! (Boss)"])
		else
			addon:Say(L["No buffs needed! (Trash)"])
		end
	end
end




function addon:ReadRaid()
	raid.readid = raid.readid + 1
	local raidnum = GetNumRaidMembers()
	local partynum = GetNumPartyMembers()
	if raidnum < 2 then
		if partynum < 1 then
			raid.reset()
			return
		else
			addon:Debug("Party detected.")
			raid.isparty = true
			raid.israid = false
			raid.size = partynum + 1
			for i = 1, partynum do
				addon:ReadUnit("party" .. i)
			end
			addon:ReadUnit("player")
		end
	else
		if raid.isparty then
			addon:Debug("Party has converted to Raid!")
			raid.reset()
		else
			addon:Debug("Raid detected.")
		end
		raid.isparty = false
		raid.israid = true
		raid.size = raidnum
		for i = 1, raidnum do
			addon:ReadUnit("raid" .. i)
		end
	end
	addon:DeleteOldUnits()
	addon:CalculateClassNumbers()
	for speccheck, _ in pairs(SP) do
		if SP[speccheck].callalways then
			SP[speccheck].code()
		end
	end
end


-- raid = { classes = { CLASS = { NAME = { readid, unitid, group, zone, online, isdead, istank, talents = {spec, tree = { talent = {} 
function addon:ReadUnit(unitid)
	if not UnitExists(unitid) then
		addon:Debug("Unit does not exist:" .. unitid)
		return
	end
	local unitindex = select(3, unitid:find("(%d+)"))
	local name = UnitName(unitid)
	if name and name ~= UNKNOWNOBJECT and name ~= UKNOWNBEING then

		local class = select(2, UnitClass(unitid))
		local isDead = UnitIsDeadOrGhost(unitid) or false
		local rank = 0
		local subgroup = 1
		local online =  UnitIsConnected(unitid)
		local role = ""
		local zone = "UNKNOWN"
		local nametwo = name
		local isML = false
		local istank = false
		local level = UnitLevel(unitid)
		
		if raid.israid then
			nametwo, rank, subgroup, _, _, _, zone, _, _, role, isML = GetRaidRosterInfo(unitindex)
			if name ~= nametwo then
				addon:Debug("NAMES DIFFER!" .. name .. " " .. nametwo)
			end
		end
		if addon.db.profile.IgnoreLastThreeGroups then
			if subgroup > 5 then
				raid.size = raid.size - 1
				return -- quick hack for now to ignore ppl in grps 6-8
			end
		end
		if raid.classes[class][name] == nil then
			raid.classes[class][name] = {}
		end
--	addon:Debug("checktank")
		if (string.find(tankList, '|' .. name .. '|')) or role == "maintank" then
	addon:Debug("intank:" .. name)
			if class ~= "PRIEST" and class ~= "ROGUE" then
				if class == "PALADIN" then
					if Aura:UnitHasBuff(unitid, BS["Righteous Fury"]) then
						istank = true
						addon:Debug("IStank:" .. name)
					end
				elseif class == "WARRIOR" then
					if Aura:UnitHasBuff(unitid, BS[7164]) then -- Defensive Stance
						istank = true
						addon:Debug("IStank:" .. name)
					end
				elseif class == "DRUID" then
					local powerType = UnitPowerType(unitid) or -1
					if powerType == 1 then -- bear form
						istank = true
						addon:Debug("IStank:" .. name)
					end
				else
					istank = true
				end
			end
		end
		
		local rcn = raid.classes[class][name]
		rcn.readid = raid.readid
		rcn.unitid = unitid
		rcn.group = subgroup
		rcn.zone = zone
		rcn.online = online
		rcn.isdead = isDead
		rcn.role = role
		rcn.rank = rank
		rcn.istank = istank
		rcn.level = level
	end
end

function addon:DeleteOldUnits()
	for class,_ in pairs(raid.classes) do
		for name,_ in pairs(raid.classes[class]) do
			if raid.classes[class][name].readid < raid.readid then
				addon:Debug("Left raid:" .. name)
				raid.classes[class][name] = nil
			end
		end
	end
end

function addon:RequestInspect(name, class, unitid)
	if raid.classes[class][name].level < 10 then
--		addon:Debug("Inspect skipping: " .. name .. " " .. raid.classes[class][name].level);
		return -- no talents below lvl 10
	end
	if _G.InspectFrame and _G.InspectFrame:IsShown() then
		return -- can't inspect at same time as UI
	end
	if CanInspect(unitid) and inspectqueuetime < GetTime() then
		if inspectqueuename == name then
			inspectqueuename = ""
			inspectqueueclass = ""
			inspectqueueunitid = ""
			return -- must be a timed out retry so skip and allow next request to get in the queue
		end
	 	addon:Debug("Inspect request going to be sent for: " .. name .. " " .. class .. " " .. unitid);
		inspectqueuetime = GetTime() + 15
		inspectqueuename = name
		inspectqueueunitid = unitid
		inspectqueueclass = class
		if UnitIsUnit(inspectqueueunitid, "player") then
			addon:INSPECT_TALENT_READY()
		else
			NotifyInspect(unitid)
			addon:RegisterEvent("INSPECT_TALENT_READY")
		end
	end
end


function addon:INSPECT_TALENT_READY()
	if addon:IsEventRegistered("INSPECT_TALENT_READY") then
		self:UnregisterEvent("INSPECT_TALENT_READY")
	end
	if inspectqueuename == "" then
--		addon:Debug("INSPECT_TALENT_READY: nothing queued");
		inspectqueuetime = 0
		return
	end
	local isnotme = true
	if UnitIsUnit(inspectqueueunitid, "player") then
--		addon:Debug("INSPECT_TALENT_READY: player is me");
		isnotme = false
	end
	if GetNumTalentTabs(isnotme) == 0 then -- sometimes the inspect fails
--		addon:Debug("Inspect failed: " .. inspectqueuename);
		ClearInspectPlayer()
		inspectqueuetime = 0
		return
	end
	local name = inspectqueuename
	local class = inspectqueueclass
	inspectqueuename = ""
	inspectqueueclass = ""
	raid.classes[class][name].talents = { tree = {}, spec = "UNKNOWN", specicon = "UNKNOWN", specialisations = {} }
	local tree = raid.classes[class][name].talents.tree
	local tabtotal = {0, 0, 0}
	for tab = 1, GetNumTalentTabs(isnotme) do
		local name, icon, spent = GetTalentTabInfo(tab, isnotme);
		tree[tab] = {name=name, icon=icon, spent=spent, talent={ }};
		for talent = 1, GetNumTalents(tab, isnotme) do
			tree[tab].talent[talent] = select(5, GetTalentInfo(tab, talent, isnotme));
			tabtotal[tab] = tabtotal[tab] + tree[tab].talent[talent]
			addon:Debug(tab .. " " .. talent .. " " .. tree[tab].talent[talent])
		end
	end
	if tabtotal[1] > 31 then
		raid.classes[class][name].talents.spec = tree[1].name
		raid.classes[class][name].talents.specicon = tree[1].icon
	elseif tabtotal[2] > 31 then
		raid.classes[class][name].talents.spec = tree[2].name
		raid.classes[class][name].talents.specicon = tree[2].icon
	elseif tabtotal[3] > 31 then
		raid.classes[class][name].talents.spec = tree[3].name
		raid.classes[class][name].talents.specicon = tree[3].icon
	else
		raid.classes[class][name].talents.spec = "Hybrid"
	end
	addon:Debug(class .. " " .. name .. " " .. raid.classes[class][name].talents.spec .. " " .. raid.classes[class][name].talents.specicon )
	addon:Debug("Inspect request finished: " .. name);
	if addon.talentframe:IsVisible() then
		addon:UpdateTalentsFrame()
	end
	inspectqueuetime = 0
end


function addon:CalculateClassNumbers()
	for _,class in ipairs(classes) do
		raid.ClassNumbers[class] = addon:DicSize(raid.classes[class])
	end
end


function addon:Say(msg, player)
	local str = "RBS::"
	local canspeak = IsRaidLeader() or IsRaidOfficer()
	for _,s in pairs({strsplit(" ", msg)}) do
		if #str + #s >= 250 then
			if player then
				SendChatMessage(str, "WHISPER", nil, player)
			else
				if addon.db.profile.ReportChat and raid.isparty then SendChatMessage(str, "party") end
				if addon.db.profile.ReportChat and raid.israid and canspeak then SendChatMessage(str, "raid") end
				if addon.db.profile.ReportSelf then addon:Print(str) end
			end
			str = "RBS::"
		end
		str = str .. " " .. s
	end
	if player then
		SendChatMessage(str, "WHISPER", nil, player)
	else
		if addon.db.profile.ReportChat and raid.isparty then SendChatMessage(str, "party") end
		if addon.db.profile.ReportChat and raid.israid and canspeak then SendChatMessage(str, "raid") end
		if addon.db.profile.ReportSelf then addon:Print(str) end
	end
end


function addon:DicSize(dic) -- is there really no built-in function to do this??
	local i = 0
	for _,_ in pairs(dic) do
		i = i + 1
	end
	return i
end


function addon:SetupFrames()
	-- main frame
	addon.frame = CreateFrame("Frame", "RBSFrame", UIParent)
	addon.frame:Hide()
	addon.frame:EnableMouse(true)
	addon.frame:SetFrameStrata("MEDIUM")
	addon.frame:SetMovable(true)
	addon.frame:SetToplevel(true)
	addon.frame:SetWidth(128)
	addon.frame:SetHeight(190)
	addon.rbsfs = addon.frame:CreateFontString("$parentTitle","ARTWORK","GameFontNormal")
	addon.rbsfs:SetText("RBS " .. GetAddOnMetadata("RaidBuffStatus", "Version"))
	addon.rbsfs:SetPoint("TOP",0,-5)
	addon.rbsfs:SetTextColor(.9,0,0)
	addon.rbsfs:Show()
	addon.frame:SetBackdrop( { 
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", 
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", tile = true, tileSize = 16, edgeSize = 16, 
		insets = { left = 5, right = 5, top = 5, bottom = 5 }
	})
	addon.frame:SetBackdropBorderColor(0,0,0,1)
	addon.frame:SetBackdropColor(0,0,0,0.99)
	addon.frame:ClearAllPoints()
	addon.frame:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
	addon.frame:SetScript("OnMouseDown",function()
		if ( arg1 == "LeftButton" ) then
			if not addon.db.profile.LockWindow then
				this:StartMoving()
			end
		end
	end)
	addon.frame:SetScript("OnMouseUp",function()
		if ( arg1 == "LeftButton" ) then
			this:StopMovingOrSizing()
			addon:SaveFramePosition()
		end
	end)
	addon.frame:SetScript("OnHide",function() this:StopMovingOrSizing() end)
	addon.frame:SetClampedToScreen(true)
	addon:LoadFramePosition()

	addon:AddBuffButtons()
	addon.button = CreateFrame("Button", "", addon.frame, "OptionsButtonTemplate")
	addon.button:SetText(L["Boss"])
	addon.button:SetWidth(45)
	addon.button:SetPoint("BOTTOMLEFT", addon.frame, "BOTTOMLEFT", 7, 5)
	addon.button:SetScript("OnClick", function() addon:DoReport() addon:ReportToChat(true) end)
	addon.button:Show()
	addon.button = CreateFrame("Button", "", addon.frame, "OptionsButtonTemplate")
	addon.button:SetText(L["Trash"])
	addon.button:SetWidth(45)
	addon.button:SetPoint("BOTTOMRIGHT", addon.frame, "BOTTOMRIGHT", -7, 5)
	addon.button:SetScript("OnClick", function() addon:DoReport() addon:ReportToChat(false) end)
	addon.button:Show()
	addon.button = CreateFrame("Button", "", addon.frame, "OptionsButtonTemplate")
	addon.button:SetText(L["R"])
	addon.button:SetWidth(22)
	addon.button:SetPoint("BOTTOM", addon.frame, "BOTTOM", 0, 5)
	addon.button:SetScript("OnClick", function()
		if IsRaidLeader() or IsRaidOfficer() then
			DoReadyCheck()
		end
	end)
	addon.button:Show()

	addon.talentsbutton = CreateFrame("Button", "talentsbutton", addon.frame, "SecureActionButtonTemplate")
	addon.talentsbutton:SetWidth(20)
	addon.talentsbutton:SetHeight(20)
	addon.talentsbutton:SetNormalTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Up")
	addon.talentsbutton:SetPushedTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Down") 
	addon.talentsbutton:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
	addon.talentsbutton:ClearAllPoints()
	addon.talentsbutton:SetPoint("TOPLEFT", addon.frame, "TOPLEFT", 5, -5)
	addon.talentsbutton:SetScript("OnClick", function()
		addon:ToggleTalentsFrame()
	end
	)
	addon.talentsbutton:Show()

	addon.optionsbutton = CreateFrame("Button", "optionsbutton", addon.frame, "SecureActionButtonTemplate")
	addon.optionsbutton:SetWidth(20)
	addon.optionsbutton:SetHeight(20)
	addon.optionsbutton:SetNormalTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Up")
	addon.optionsbutton:SetPushedTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Down") 
	addon.optionsbutton:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
	addon.optionsbutton:ClearAllPoints()
	addon.optionsbutton:SetPoint("TOPRIGHT", addon.frame, "TOPRIGHT", -5, -5)
	addon.optionsbutton:SetScript("OnClick", function()
		addon:ToggleOptionsFrame()
	end
	)
	addon.optionsbutton:Show()

	-- talents window frame

	addon.talentframe = CreateFrame("Frame", "RBSTalentsFrame", UIParent, "DialogBoxFrame")
	addon.talentframe:Hide()
	addon.talentframe:EnableMouse(true)
	addon.talentframe:SetFrameStrata("MEDIUM")
	addon.talentframe:SetMovable(true)
	addon.talentframe:SetToplevel(true)
	addon.talentframe:SetWidth(tfi.framewidth)
	addon.talentframe:SetHeight(190)
	addon.rbsfs = addon.talentframe:CreateFontString("$parentTitle","ARTWORK","GameFontNormal")
	addon.rbsfs:SetText("RBS " .. GetAddOnMetadata("RaidBuffStatus", "Version") .. " - " .. L["Talent Specialisations"])
	addon.rbsfs:SetPoint("TOP",0,-5)
	addon.rbsfs:SetTextColor(.9,0,0)
	addon.rbsfs:Show()
	addon.talentframe:SetBackdrop( { 
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", 
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", tile = true, tileSize = 16, edgeSize = 16, 
		insets = { left = 5, right = 5, top = 5, bottom = 5 }
	})
	addon.talentframe:SetBackdropBorderColor(0,0,0,1)
	addon.talentframe:SetBackdropColor(0,0,0,0.99)
	addon.talentframe:ClearAllPoints()
	addon.talentframe:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
	addon.talentframe:SetScript("OnMouseDown",function()
		if ( arg1 == "LeftButton" ) then
			if not addon.db.profile.LockWindow then
				this:StartMoving()
			end
		end
	end)
	addon.talentframe:SetScript("OnMouseUp",function()
		if ( arg1 == "LeftButton" ) then
			this:StopMovingOrSizing()
			addon:SaveFramePosition()
		end
	end)
	addon.talentframe:SetScript("OnHide",function() this:StopMovingOrSizing() end)
	addon.button = CreateFrame("Button", "", addon.talentframe, "OptionsButtonTemplate")
	addon.button:SetText(L["Name"])
	addon.button:SetWidth(tfi.namewidth)
	addon.button:SetPoint("TOPLEFT", addon.talentframe, "TOPLEFT", tfi.namex, -20)
	addon.button:SetScript("OnClick", function()
		tfi.sort = "name"
		tfi.sortorder = 0 - tfi.sortorder
		addon:ShowTalentsFrame()
	end)
	addon.button:Show()
	addon.button = CreateFrame("Button", "", addon.talentframe, "OptionsButtonTemplate")
	addon.button:SetText(L["Class"])
	addon.button:SetWidth(tfi.classwidth)
	addon.button:SetPoint("TOPLEFT", addon.talentframe, "TOPLEFT", tfi.classx, -20)
		addon.button:SetScript("OnClick", function()
		tfi.sort = "class"
		tfi.sortorder = 0 - tfi.sortorder
		addon:ShowTalentsFrame()
	end)
	addon.button:Show()
	addon.button = CreateFrame("Button", "", addon.talentframe, "OptionsButtonTemplate")
	addon.button:SetText(L["Spec"])
	addon.button:SetWidth(tfi.specwidth)
	addon.button:SetPoint("TOPLEFT", addon.talentframe, "TOPLEFT", tfi.specx, -20)
		addon.button:SetScript("OnClick", function()
		tfi.sort = "spec"
		tfi.sortorder = 0 - tfi.sortorder
		addon:ShowTalentsFrame()
	end)
	addon.button:Show()
	addon.button = CreateFrame("Button", "", addon.talentframe, "OptionsButtonTemplate")
	addon.button:SetText(L["Specialisations"])
	addon.button:SetWidth(tfi.specialisationswidth)
	addon.button:SetPoint("TOPLEFT", addon.talentframe, "TOPLEFT", tfi.specialisationsx, -20)
		addon.button:SetScript("OnClick", function()
		tfi.sort = "specialisations"
		tfi.sortorder = 0 - tfi.sortorder
		addon:ShowTalentsFrame()
	end)
	addon.button:Show()

	rowy = 0 - tfi.topedge
	for i = 1, tfi.maxrows do
		tfi.rowframes[i] = {}
		addon.rowframe = CreateFrame("Frame", "", addon.talentframe)
		tfi.rowframes[i].rowframe = addon.rowframe
		addon.rowframe:SetWidth(tfi.rowwidth)
		addon.rowframe:SetHeight(tfi.rowheight)
		addon.rowframe:ClearAllPoints()
		addon.rowframe:SetPoint("TOPLEFT", addon.talentframe, "TOPLEFT", tfi.edge + tfi.inset, rowy)
		addon.rbsfs = addon.rowframe:CreateFontString(nil,"ARTWORK","GameFontNormal")
		tfi.rowframes[i].name = addon.rbsfs
		addon.rbsfs:SetText("Whoami")
		addon.rbsfs:SetPoint("TOPLEFT", addon.rowframe, "TOPLEFT", 0, -2)
		addon.rbsfs:SetTextColor(.9,0,0)
		addon.rbsfs:Show()

		addon.button = CreateFrame("Button", "", addon.rowframe)
		tfi.rowframes[i].class = addon.button
		addon.button:SetWidth(tfi.buttonsize)
		addon.button:SetHeight(tfi.buttonsize)
		addon.button:SetNormalTexture("Interface\\Icons\\INV_ValentinesCandy")
		addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
		addon.button:SetPoint("TOPLEFT", addon.rowframe, "TOPLEFT", tfi.classx + ((tfi.classwidth - 30) / 2), 0)
		addon.button:Show()

		addon.button = CreateFrame("Button", "", addon.rowframe)
		tfi.rowframes[i].spec = addon.button
		addon.button:SetWidth(tfi.buttonsize)
		addon.button:SetHeight(tfi.buttonsize)
		addon.button:SetNormalTexture("Interface\\Icons\\Ability_ThunderBolt")
		addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
		addon.button:SetPoint("TOPLEFT", addon.rowframe, "TOPLEFT", tfi.specx + ((tfi.specwidth - 30) / 2), 0)
		addon.button:Show()
		
		tfi.rowframes[i].specialisations = {}
		addon.button = CreateFrame("Button", "", addon.rowframe)
		tfi.rowframes[i].specialisations[1] = addon.button
		addon.button:SetWidth(tfi.buttonsize)
		addon.button:SetHeight(tfi.buttonsize)
		addon.button:SetNormalTexture("Interface\\Icons\\Ability_ThunderBolt")
		addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
		addon.button:SetPoint("TOPRIGHT", addon.rowframe, "TOPRIGHT", 0 - tfi.inset, 0)
		addon.button:Show()
		addon.button = CreateFrame("Button", "", addon.rowframe)
		tfi.rowframes[i].specialisations[2] = addon.button
		addon.button:SetWidth(tfi.buttonsize)
		addon.button:SetHeight(tfi.buttonsize)
		addon.button:SetNormalTexture("Interface\\Icons\\Ability_ThunderBolt")
		addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
		addon.button:SetPoint("TOPRIGHT", tfi.rowframes[i].specialisations[1], "TOPLEFT", 0 - tfi.gap, 0)
		addon.button:Show()
		addon.button = CreateFrame("Button", "", addon.rowframe)
		tfi.rowframes[i].specialisations[3] = addon.button
		addon.button:SetWidth(tfi.buttonsize)
		addon.button:SetHeight(tfi.buttonsize)
		addon.button:SetNormalTexture("Interface\\Icons\\Ability_ThunderBolt")
		addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
		addon.button:SetPoint("TOPRIGHT", tfi.rowframes[i].specialisations[2], "TOPLEFT", 0 - tfi.gap, 0)
		addon.button:Show()
		addon.button = CreateFrame("Button", "", addon.rowframe)
		tfi.rowframes[i].specialisations[4] = addon.button
		addon.button:SetWidth(tfi.buttonsize)
		addon.button:SetHeight(tfi.buttonsize)
		addon.button:SetNormalTexture("Interface\\Icons\\Ability_ThunderBolt")
		addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
		addon.button:SetPoint("TOPRIGHT", tfi.rowframes[i].specialisations[3], "TOPLEFT", 0 - tfi.gap, 0)
		addon.button:Show()
		addon.button = CreateFrame("Button", "", addon.rowframe)
		tfi.rowframes[i].specialisations[5] = addon.button
		addon.button:SetWidth(tfi.buttonsize)
		addon.button:SetHeight(tfi.buttonsize)
		addon.button:SetNormalTexture("Interface\\Icons\\Ability_ThunderBolt")
		addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
		addon.button:SetPoint("TOPRIGHT", tfi.rowframes[i].specialisations[4], "TOPLEFT", 0 - tfi.gap, 0)
		addon.button:Show()
		addon.button = CreateFrame("Button", "", addon.rowframe)
		tfi.rowframes[i].specialisations[6] = addon.button
		addon.button:SetWidth(tfi.buttonsize)
		addon.button:SetHeight(tfi.buttonsize)
		addon.button:SetNormalTexture("Interface\\Icons\\Ability_ThunderBolt")
		addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
		addon.button:SetPoint("TOPRIGHT", tfi.rowframes[i].specialisations[5], "TOPLEFT", 0 - tfi.gap, 0)
		addon.button:Show()
		rowy = rowy - tfi.rowheight - tfi.rowgap
	end


	-- options window frame
	addon.optionsframe = CreateFrame("Frame", "RBSOptionsFrame", UIParent, "DialogBoxFrame")
	addon.optionsframe:Hide()
	addon.optionsframe:EnableMouse(true)
	addon.optionsframe:SetFrameStrata("MEDIUM")
	addon.optionsframe:SetMovable(true)
	addon.optionsframe:SetToplevel(true)
	addon.optionsframe:SetWidth(300)
	addon.optionsframe:SetHeight(228)
	addon.rbsfs = addon.optionsframe:CreateFontString("$parentTitle","ARTWORK","GameFontNormal")
	addon.rbsfs:SetText("RBS " .. GetAddOnMetadata("RaidBuffStatus", "Version") .. " - " .. L["Options"])
	addon.rbsfs:SetPoint("TOP",0,-5)
	addon.rbsfs:SetTextColor(.9,0,0)
	addon.rbsfs:Show()
	addon.optionsframe:SetBackdrop( { 
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", 
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", tile = true, tileSize = 16, edgeSize = 16, 
		insets = { left = 5, right = 5, top = 5, bottom = 5 }
	})
	addon.optionsframe:SetBackdropBorderColor(0,0,0,1)
	addon.optionsframe:SetBackdropColor(0,0,0,0.99)
	addon.optionsframe:ClearAllPoints()
	addon.optionsframe:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
	addon.optionsframe:SetScript("OnMouseDown",function()
		if ( arg1 == "LeftButton" ) then
			if not addon.db.profile.LockWindow then
				this:StartMoving()
			end
		end
	end)
	addon.optionsframe:SetScript("OnMouseUp",function()
		if ( arg1 == "LeftButton" ) then
			this:StopMovingOrSizing()
			addon:SaveFramePosition()
		end
	end)
	addon.optionsframe:SetScript("OnHide",function() this:StopMovingOrSizing() end)

-- incomplete future slider option for width control
--	addon.optionsframe.slider = CreateFrame("Slider", nil, addon.optionsframe)
--	addon.optionsframe.slider:SetOrientation("HORIZONTAL")
--	addon.optionsframe.slider:SetMinMaxValues(5, 10)
--	addon.optionsframe.slider:SetValueStep(1)
--	addon.optionsframe.slider:SetValue(5)
--	addon.optionsframe.slider:SetWidth(8)
--	addon.optionsframe.slider:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
--	addon.optionsframe.slider:SetBackdrop({
--		bgFile = "Interface\\Buttons\\UI-SliderBar-Background",
--		edgeFile = "Interface\\Buttons\\UI-SliderBar-Border",
--		tile = true, tileSize = 8, edgeSize = 8,
--		insets = { left = 3, right = 3, top = 6, bottom = 6, } })
--	addon.optionsframe.slider:SetThumbTexture("Interface\\Buttons\\UI-SliderBar-Button-Horizontal")
--	addon.optionsframe.slider:Show()

	addon.rbsfs = addon.optionsframe:CreateFontString("$parentTitle","ARTWORK","GameFontNormal")
	addon.rbsfs:SetText(L["Is a warning"] .. ":")
	addon.rbsfs:SetPoint("TOPLEFT",10,-53)
	addon.rbsfs:SetTextColor(.9,0,0)
	addon.rbsfs:Show()
	addon.rbsfs = addon.optionsframe:CreateFontString("$parentTitle","ARTWORK","GameFontNormal")
	addon.rbsfs:SetText(L["Is a buff"] .. ":")
	addon.rbsfs:SetPoint("TOPLEFT",10,-73)
	addon.rbsfs:SetTextColor(.9,0,0)
	addon.rbsfs:Show()
	addon.rbsfs = addon.optionsframe:CreateFontString("$parentTitle","ARTWORK","GameFontNormal")
	addon.rbsfs:SetText(L["Show on dashboard"] .. ":")
	addon.rbsfs:SetPoint("TOPLEFT",10,-93)
	addon.rbsfs:SetTextColor(.9,0,0)
	addon.rbsfs:Show()
	addon.rbsfs = addon.optionsframe:CreateFontString("$parentTitle","ARTWORK","GameFontNormal")
	addon.rbsfs:SetText(L["Show in combat"] .. ":")
	addon.rbsfs:SetPoint("TOPLEFT",10,-113)
	addon.rbsfs:SetTextColor(.9,0,0)
	addon.rbsfs:Show()
	addon.rbsfs = addon.optionsframe:CreateFontString("$parentTitle","ARTWORK","GameFontNormal")
	addon.rbsfs:SetText(L["Report on Trash"] .. ":")
	addon.rbsfs:SetPoint("TOPLEFT",10,-133)
	addon.rbsfs:SetTextColor(.9,0,0)
	addon.rbsfs:Show()
	addon.rbsfs = addon.optionsframe:CreateFontString("$parentTitle","ARTWORK","GameFontNormal")
	addon.rbsfs:SetText(L["Report on Boss"] .. ":")
	addon.rbsfs:SetPoint("TOPLEFT",10,-153)
	addon.rbsfs:SetTextColor(.9,0,0)
	addon.rbsfs:Show()

	local bufflist = {}
	for buffcheck, _ in pairs(BF) do
		table.insert(bufflist, buffcheck)
	end
	table.sort(bufflist, function (a,b) return BF[a].order > BF[b].order end)

	local saveradio = function ()
		local name = this:GetName()
		addon.db.profile[name] = this:GetChecked() and true or false
		local buffradio = false
		local isbuff = false
		if name:find("buff$") then
			buffradio = name:sub(1, name:find("buff$") - 1)
			isbuff = true
		elseif name:find("warning$") then
			buffradio = name:sub(1, name:find("warning$") - 1)
		end
		if buffradio then
			local value = true
			if addon.db.profile[name] then
				value = false  -- if I am ticked then make the other unticked
			end
			if isbuff then
				addon.db.profile[buffradio .. "warning"] = value
			else
				addon.db.profile[buffradio .. "buff"] = value
			end
			addon:UpdateOptionsButtons()
		end
		addon:AddBuffButtons()
		addon:UpdateButtons()
	end

	local currentx = 140
	for _, buffcheck in ipairs(bufflist) do
		addon:AddOptionsBuffButton(buffcheck, currentx, -25, BF[buffcheck].icon, BF[buffcheck].tip)
		addon:AddOptionsBuffRadioButton(buffcheck .. "warning", currentx, -50, saveradio)
		addon:AddOptionsBuffRadioButton(buffcheck .. "buff", currentx, -70, saveradio)
		addon:AddOptionsBuffRadioButton(buffcheck .. "dash", currentx, -90, saveradio)
		addon:AddOptionsBuffRadioButton(buffcheck .. "dashcombat", currentx, -110, saveradio)
		addon:AddOptionsBuffRadioButton(buffcheck .. "trash", currentx, -130, saveradio)
		addon:AddOptionsBuffRadioButton(buffcheck .. "boss", currentx, -150, saveradio)
		currentx = currentx + 22
	end
	addon.optionsframe:SetWidth(currentx + 9)
end


function addon:SaveFramePosition()
	addon.db.profile.x = addon.frame:GetLeft()
	addon.db.profile.y = addon.frame:GetTop() - UIParent:GetTop()
end

function addon:LoadFramePosition()
	addon.frame:ClearAllPoints()
	if (addon.db.profile.x ~= 0) or (addon.db.profile.y ~= 0) then
		addon.frame:SetPoint("TOPLEFT", UIParent,"TOPLEFT", addon.db.profile.x, addon.db.profile.y)
	else
		addon.frame:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
	end
end


function addon:AddBuffButtons()
	local buffs = {}
	local warnings = {}
	local bosses = {}
	for _, v in ipairs(buttons) do
		v.free = true
		v:Hide()
	end
	for buffcheck, _ in pairs(BF) do
		if  (not incombat and addon.db.profile[buffcheck .. "dash"]) or (incombat and addon.db.profile[buffcheck .. "dashcombat"]) then
			if addon.db.profile[buffcheck .. "boss"] and (not addon.db.profile[buffcheck .. "trash"]) then
				table.insert(bosses, buffcheck)
			else
				if addon.db.profile[buffcheck .. "warning"] then
					table.insert(warnings, buffcheck)
				end
				if addon.db.profile[buffcheck .. "buff"] then
					table.insert(buffs, buffcheck)
				end
			end
		end
		
	end
	addon:SortButtons(bosses)
	addon:SortButtons(buffs)
	addon:SortButtons(warnings)
	local currenty = 8
	local maxcols = 5
	local cols = { 10, 32, 54, 76, 98, 120, 142, 164, 186 }
	if # bosses > 0 then
		currenty = addon:AddButtonType(bosses, maxcols, cols, currenty)
	end
	if # buffs > 0 then
		currenty = addon:AddButtonType(buffs, maxcols, cols, currenty)
	end
	if # warnings > 0 then
		currenty = addon:AddButtonType(warnings, maxcols, cols, currenty)
	end
	addon.frame:SetHeight(currenty + 42)
end

function addon:AddButtonType(buttonlist, maxcols, cols, currenty)
	for i, v in ipairs(buttonlist) do
		local x = cols[((i - 1) % maxcols) + 1]
		local y = currenty + (22 * (math.ceil((# buttonlist)/maxcols) - math.floor((i - 1) / maxcols)))
		addon:AddBuffButton(v, x, y, BF[v].icon, BF[v].update, BF[v].click, BF[v].tip)
	end
	return (currenty + 6 + 22 * math.ceil((# buttonlist)/maxcols))
end

function addon:SortButtons(buttonlist)
	table.sort(buttonlist, function (a,b)
		return (BF[a].order > BF[b].order)
	end)
end


function addon:AddBuffButton(name, x, y, icon, update, click, tooltip)
	addon.button = nil
	for _, v in ipairs(buttons) do
		if v.free then
			addon.button = v
			addon.button.update = nil
			addon.button:SetScript("PreClick", nil)
			addon.button:SetScript("PostClick", nil)
			addon.button:SetScript("OnEnter", nil)
			addon.button:SetScript("OnLeave", nil)
			addon.button:SetAttribute("type", nil)
			addon.button:SetAttribute("spell", nil)
			addon.button:SetAttribute("unit", nil)
			break
		end
	end
	if not addon.button then
		addon.button = CreateFrame("Button", name, addon.frame, "SecureActionButtonTemplate")
		table.insert(buttons, addon.button)
		addon.button:Hide()
		addon.button:SetWidth(20)
		addon.button:SetHeight(20)
		addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
		addon.button:SetAlpha(1)
		addon.count = addon.button:CreateFontString(nil, "ARTWORK","GameFontNormalSmall")
		addon.button.count = addon.count
		addon.count:SetWidth(20)
		addon.count:SetHeight(20)
		addon.count:SetFont(addon.count:GetFont(),12,"OUTLINE")
		addon.count:SetPoint("CENTER", self.button, "CENTER", 0, 0)
		addon.count:SetText("X")
		addon.count:Show()
	end
	addon.button.free = false
	addon.button:SetNormalTexture(icon)
	addon.button:SetPoint("BOTTOMLEFT", addon.frame, "BOTTOMLEFT", x, y)
	if click then
		addon.button:SetScript("PreClick", click)
	end
	if update then
		addon.button.update = update
	end
	if tooltip then
		addon.button:SetScript("OnEnter", tooltip)
		addon.button:SetScript("OnLeave", function() GameTooltip:Hide() end)
	end
	addon.button:Show()
end

function addon:AddOptionsBuffButton(name, x, y, icon, tooltip)
	addon.button = CreateFrame("Button", name, addon.optionsframe, "SecureActionButtonTemplate")
	addon.button:Hide()
	addon.button:SetWidth(20)
	addon.button:SetHeight(20)
	addon.button:SetNormalTexture(icon)
	addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
	addon.button:SetPoint("TOPLEFT", addon.optionsframe, "TOPLEFT", x, y)
	if tooltip then
		addon.button:SetScript("OnEnter", tooltip)
		addon.button:SetScript("OnLeave", function() GameTooltip:Hide() end)
	end
	addon.button:Show()
end

function addon:AddOptionsBuffRadioButton(name, x, y, click)
	addon.button = CreateFrame("CheckButton", name, addon.optionsframe, "UICheckButtonTemplate")
	table.insert(optionsbuttons, addon.button)
	addon.button:Hide()
	if id then
		addon.button:SetID(id)
	end
	addon.button:SetWidth(20)
	addon.button:SetHeight(20)
	addon.button:SetPoint("TOPLEFT", addon.optionsframe, "TOPLEFT", x, y)
	addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
	if click then
		addon.button:SetScript("OnClick", click)
	end
	addon.button:Show()
end



function addon:ToggleCheck(...)
	if addon.db.profile[...] then
		addon.db.profile[...] = false
	else
		addon.db.profile[...] = true
		addon:DoReport()
	end
	addon:UpdateButtons()
end


function addon:UpdateButtons()
	for _,v in ipairs(buttons) do
		if not v.free then
			if v.update then
				v:update(v)
			end
		end
	end
end

function addon:UpdateOptionsButtons()
	for _,v in ipairs(optionsbuttons) do
		v:SetChecked(addon.db.profile[v:GetName()])
	end
end


function addon:OnProfileEnable()
	addon:Debug("Profile changed")
	addon:LoadFramePosition()
	addon:DoReport()
end

function addon:JoinedPartyRaidChanged()
	addon:Debug("JoinedPartyRaidChanged");
	local oldstatus = raid.isparty or raid.israid
	addon:DoReport()
	if oldstatus then  -- was a raid or party last check
		if raid.isparty or raid.israid then -- still is a raid or party
			addon:Debug("Still in party or raid");
		else	-- no longer in raid or party
			addon:HideReportFrame()
			if addon:IsEventRegistered("PLAYER_AURAS_CHANGED") then
				addon:UnregisterEvent("PLAYER_AURAS_CHANGED")
			end
			if addon:IsEventScheduled("DoReport") then
				addon:CancelScheduledEvent("DoReport")
			end
		end
	else
		if raid.isparty or raid.israid then -- newly entered raid or party
			addon:ShowReportFrame()
			addon:RegisterEvent("PLAYER_AURAS_CHANGED", "DoReport")
			addon:ScheduleRepeatingEvent("DoReport", addon.DoReport, 10, addon)
		end
	end
end

function addon:NotifyInspect(unitid)
	if unitid ~= inspectqueueunitid then
--		addon:Debug("NotifyInspect mismatch:" .. unitid .. " " .. inspectqueueunitid)
		if addon:IsEventRegistered("INSPECT_TALENT_READY") then
			self:UnregisterEvent("INSPECT_TALENT_READY")
		end
		inspectqueuename = ""
		inspectqueueunitid = ""
		inspectqueuetime = 0
		return
	end
end

function addon:OnEnable()
--	addon:Print("RBS: Activated");
	addon:SetupFrames()
	hooksecurefunc("NotifyInspect", function(...) return addon:NotifyInspect(...) end)
	addon:JoinedPartyRaidChanged()
	addon:RegisterEvent("RAID_ROSTER_UPDATE", "JoinedPartyRaidChanged")
	addon:RegisterEvent("PARTY_MEMBERS_CHANGED", "JoinedPartyRaidChanged")
	addon:RegisterEvent("PLAYER_ENTERING_WORLD", "JoinedPartyRaidChanged")
	addon:RegisterEvent("PLAYER_REGEN_ENABLED", "LeftCombat")
	addon:RegisterEvent("PLAYER_REGEN_DISABLED", "EnteringCombat")
end


function addon:OnDisable()
	if addon:IsEventRegistered("RAID_ROSTER_UPDATE") then
		addon:UnregisterEvent("RAID_ROSTER_UPDATE")
	end
	if addon:IsEventRegistered("PARTY_ROSTER_UPDATE") then
		addon:UnregisterEvent("PARTY_ROSTER_UPDATE")
	end
	if addon:IsEventRegistered("PLAYER_ENTERING_WORLD") then
		addon:UnregisterEvent("PLAYER_ENTERING_WORLD")
	end
	if addon:IsEventScheduled("DoReport") then
		addon:CancelScheduledEvent("DoReport")
	end
	if addon:IsEventRegistered("INSPECT_TALENT_READY") then
		self:UnregisterEvent("INSPECT_TALENT_READY")
	end
	if addon:IsEventRegistered("PLAYER_REGEN_ENABLED") then
		self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	end
	if addon:IsEventRegistered("PLAYER_REGEN_DISABLED") then
		self:UnregisterEvent("PLAYER_REGEN_DISABLED")
	end
end

function addon:EnteringCombat()
	incombat = true
	addon:AddBuffButtons()
	addon:UpdateButtons()
end

function addon:LeftCombat()
	incombat = false
	addon:AddBuffButtons()
	addon:UpdateButtons()
end

function addon:Tooltip(self, title, list, tlist, blist)
	GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
	GameTooltip:SetText(title,1,1,1,1,1)
	if list then
		if #list > 0 then
			local thelisttouse = list
			local timerlist = {}
			if tlist then
				for _, n in ipairs(list) do
					if tlist[n] then
						table.insert(timerlist, n .. "(" .. addon:TimeSince(tlist[n]) .. ")")
					else
						table.insert(timerlist, n)
					end
				end
				thelisttouse = timerlist
			end
			local str = ""
			for _,s in pairs({strsplit(" ", table.concat(thelisttouse, ", "))}) do
				if #str + #s >= 50 then
					GameTooltip:AddLine(str,nil,nil,nil,nil,1)
					str = ""
				end
				str = str .. " " .. s
			end
			GameTooltip:AddLine(str,nil,nil,nil,nil,1)
			if blist then
				str = L["Buffers: "]
				for _,s in pairs({strsplit(" ", table.concat(blist, ", "))}) do
					if #str + #s >= 50 then
						GameTooltip:AddLine(str,nil,nil,nil,nil,1)
						str = ""
					end
					str = str .. " " .. s
				end
				GameTooltip:AddLine(str,nil,nil,nil,nil,1)
			end
		end
	end
	GameTooltip:Show()
end

function addon:DefaultButtonUpdate(self, table, profile, checking)
	if profile == false then
		self.count:SetText("X")
		self:SetAlpha("0.5")
	else
		if checking then
			self.count:SetText(# table)
			self:SetAlpha("1")
		else
			self.count:SetText("")
			self:SetAlpha("0.15")
		end
	end
end


function addon:ButtonClick(self, button, down, buffcheck, cheapspell, reagentspell, partybuff)
	local check = BF[buffcheck].check
	local prefix
	if addon.db.profile[buffcheck .. "buff"] then
		prefix = L["Missing buff: "]
	else
		prefix = L["Warning: "]
	end
	if addon.db.profile[check] then
		if cheapspell then
			self:SetAttribute("type", nil)
			self:SetAttribute("spell", nil)
			self:SetAttribute("unit", nil)
			self:SetScript("PostClick", nil)
		end
		if cheapspell and IsAltKeyDown() then
			addon:DoReport()
			if not InCombatLockdown() and # report[BF[buffcheck].list] > 0 then
				self:SetAttribute("type", "spell")
				if partybuff then
					if # report[BF[buffcheck].list] > 0 then
						local unitidtobuff, spelltobuff = addon:PartyBuff(report[BF[buffcheck].list], cheapspell, reagentspell)
						self:SetAttribute("spell", spelltobuff)
						self:SetAttribute("unit", unitidtobuff)
						if unitidtobuff then  -- maybe none in range
							self:SetScript("PostClick", function(self)
								self:SetAttribute("type", nil)
								self:SetAttribute("spell", nil)
								self:SetAttribute("unit", nil)
								self:SetScript("PostClick", nil)
							end)
						end
					end
				else
					self:SetAttribute("spell", cheapspell)
				end
			end
		elseif IsShiftKeyDown() then
			if type(BF[buffcheck].chat) == "string" then
				if # report[BF[buffcheck].list] > 0 then
					if BF[buffcheck].timer then
						local timerlist = {}
						for _, n in ipairs(report[BF[buffcheck].list]) do
							if raid.BuffTimers[buffcheck .. "timerlist"][n] then
								table.insert(timerlist, n .. "(" .. addon:TimeSince(raid.BuffTimers[buffcheck .. "timerlist"][n]) .. ")")
							else
								table.insert(timerlist, n)
							end
						end
						addon:Say("<" .. BF[buffcheck].chat .. ">: " .. table.concat(timerlist, ", "))
					else
						addon:Say(prefix .. "<" .. BF[buffcheck].chat .. ">: " .. table.concat(report[BF[buffcheck].list], ", "))
					end
				end
			elseif type(BF[buffcheck].chat) == "function" then
				BF[buffcheck].chat(report, raid, prefix)
			end
		elseif IsControlKeyDown() then
			if BF[buffcheck].selfbuff then
				if type(BF[buffcheck].chat) == "string" then
					if # report[BF[buffcheck].list] > 0 then
						for _, v in ipairs(report[BF[buffcheck].list]) do
							local name = v
							if v:find("%(") then
								name = string.sub(v, 1, v:find("%(") - 1)
							end
							addon:Say(prefix .. "<" .. BF[buffcheck].chat .. ">: " .. v, name)
						end
					end
				elseif type(BF[buffcheck].chat) == "function" then
					BF[buffcheck].chat(report, raid, prefix)
				end
			elseif BF[buffcheck].partybuff then
				BF[buffcheck].partybuff(report[BF[buffcheck].list], prefix)
			end
		else
			addon:ToggleCheck(check)
		end
	else
		addon:ToggleCheck(check)
	end
end

function addon:SortNameBySuffix(thetable)
	table.sort(thetable, function (a,b)
		local grpa = select(3, a:find"(%d+)")
		local grpb = select(3, b:find"(%d+)")
		if grpa and grpb then
			if grpa == grpb then
				return a < b
			end
			return grpa < grpb
		else
			return a < b
		end
	end)
end

function addon:TitleCaps(str)
	str = string.lower(str)
	return str:gsub("%a", string.upper, 1)
end

function addon:TimeSince(thetimethen)
	local thedifference = math.floor(GetTime() - thetimethen)
	if thedifference < 60 then
		return (thedifference .. "s")
	end
	return (math.floor(thedifference / 60) .. "m" .. (thedifference % 60) .. "s")
end

function addon:InGroupWithEnhShaman(unit)
	for name,_ in pairs(raid.classes.SHAMAN) do
		if raid.classes.SHAMAN[name].talents then
			if raid.classes.SHAMAN[name].talents.spec == "Enhancement" then
				if raid.classes.SHAMAN[name].group == unit.group then
					return true
				end
			end
		end
	end
	return false
end

function addon:IsMele(name, class)
	if class == "ROGUE" then return true end
	if class == "WARRIOR" then return true end
	if class == "PALADIN" then
		if raid.classes.PALADIN[name].talents then
			if raid.classes.PALADIN[name].talents.spec ~= "Holy" then
				return true
			end
		end
		return false
	end
	if class == "DRUID" then
		if raid.classes.DRUID[name].talents then
			if raid.classes.DRUID[name].talents.spec == "Feral Combat" then
				return true
			end
		end
	end
	if class == "SHAMAN" then
		if raid.classes.SHAMAN[name].talents then
			if raid.classes.SHAMAN[name].talents.spec == "Enhancement" then
				return true
			end
		end
	end
	return false
end

function addon:NumberOfBlessings(unitid, nosalv)
	local list = blessings
	if nosalv then
		list = blessingsns
	end
	local numbless = 0
	for _, v in ipairs(list) do
		if Aura:UnitHasBuff(unitid, v) then
			numbless = numbless + 1
		end
	end
	return numbless
end




function addon:GetUnitFromName(whom)
	if whom:find("%(") then
		whom = string.sub(whom, 1, whom:find("%(") - 1)
	end
	for class,_ in pairs(raid.classes) do
		for name,_ in pairs(raid.classes[class]) do
			if whom == name then
				return raid.classes[class][name]
			end
		end
	end
	return nil
end


function addon:PartyBuff(list, cheapspell, reagentspell)
	local pb = {{},{},{},{},{},{},{},{}}
	for _, v in ipairs(list) do
		if v:find("%(") then
			v = string.sub(v, 1, v:find("%(") - 1)
		end
		local unit = addon:GetUnitFromName(v)
		table.insert(pb[unit.group], unit.unitid)
	end
	table.sort(pb, function (a,b)
		return(#a > #b)
	end)
	for i,_ in ipairs(pb) do
		for _, v in ipairs(pb[i]) do
			if IsSpellInRange(cheapspell, v) == 1 then
				if #pb[i] > 2 then
					return v, reagentspell
				end
				return v, cheapspell
			end
		end
	end
	addon:Print("RBS: " .. L["Out of range"])
	return nil
end
